#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Coyote - Advanced Pilot Training Aircraft
# Plane designed for the AIAA 2017-2018 aircraft design competition
# by Raphael Dubois, Thibault Laurent, Bao Long Le Van, Nayan Levaux,
#    Guillaume Noiset, Axel Piret, Arthur Scheffer, and Vincent Schmitz
# CAD created by Nayan Levaux and exported by Adrien Crovato
# Valid mesh obtained with gmsh 4.4.1

def getParam():
    p = {}
    # Specific
    scl = 2 # scaling factor for lifting surface mesh size
    wrtems = scl*0.036 # wing root trailing edge mesh size
    trtems = scl*0.028 # tail root trailing mesh size
    fums = 0.1 # fuselage mesh size
    fams = 5. # farfield mesh size
    # Input/Output
    p['File'] = '../models/coyote.geo' # Input file containing the model
    p['Pars'] = {'msWing' : wrtems, 'msTail' : trtems,
                 'msFus' : fums, 'msFar' : fams} # Parameters for input file model
    p['Dim'] = 3 # Problem dimension
    p['Format'] = 'vtk' # Save format (vtk or gmsh)
    p['Slice'] = [1.8, 3.5, 5.3] # array of (y) coordinates to perform slice along the span (empty if none)
    p['TagId'] = 9 # tag number of physical group to be sliced
    # Markers
    p['Fluid'] = 'field' # Name of physical group containing the fluid
    p['Farfield'] = ['upstream', 'farfield', 'downstream'] # LIST of names of physical groups containing the farfield boundaries (upstream/downstream should be first/last element)
    p['Fuselage'] = 'fuselage' # Name of physical group containing the fuselage boundaries
    p['Wings'] = ['wing', 'tail'] # LIST of names of physical groups containing the lifting surface boundary
    p['Wakes'] = ['wakeW', 'wakeT'] # LIST of names of physical group containing the wake
    p['WakeTips'] = ['wakeTipW', 'wakeTipT'] # LIST of names of physical group containing the edge of the wake
    p['TeTips'] = ['teTipW', 'teTipT'] # LIST of names of physical group containing the edge of the wake and the trailing edge
    # Freestream
    p['M_inf'] = 0.8 # Freestream Mach number
    p['AoA_begin'] = 1 # Freestream angle of attack (begin)
    p['AoA_end'] = 1 # Freestream angle of attack (end)
    p['AoA_step'] = 1. # Freestream angle of attack (step)
    p['AoS'] = 3. # Freestream angle of sideslip
    # Geometry
    p['S_ref'] = 25 # Reference surface length
    p['c_ref'] = 2.6 # Reference chord length
    p['x_ref'] = 8.9 # Reference point for moment computation (x)
    p['y_ref'] = 1.8 # Reference point for moment computation (y)
    p['z_ref'] = 0.8 # Reference point for moment computation (z)
    # Numerical
    p['LSolver'] = 'GMRES' # Linear solver (Pardiso, GMRES, MUMPS or SparseLU)
    p['G_fill'] = 2 # Fill-in factor for GMRES preconditioner
    p['G_tol'] = 1e-5 # Tolerance for GMRES
    p['G_restart'] = 50 # Restart for GMRES
    p['NSolver'] = 'Newton' # Noninear solver type (Picard or Newton)
    p['Rel_tol'] = 1e-6 # Relative tolerance on solver residual
    p['Abs_tol'] = 1e-8 # Absolute tolerance on solver residual
    p['Max_it'] = 15 # Solver maximum number of iterations
    p['LS_tol'] = 1e-6 # Tolerance on linesearch residual
    p['Max_it_LS'] = 10 # Linesearch maximum number of iterations
    p['AV_thrsh'] = 1e-2 # Residual threshold below which the artificial viscosity is decreased
    p['M_crit'] = 5. # Critical Mach number above which density is damped
    return p

def main():
    # Script call
    import flow.scripts.polar as p
    polar = p.Polar(getParam())
    polar.run()
    polar.disp()

if __name__ == "__main__":
    main()
