/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wAssign.h"
#include "wProblem.h"
#include "wTag.h"
#include "wElement.h"
#include "wNode.h"
#include <algorithm>
using namespace tbox;
using namespace flow;

Assign::Assign(std::shared_ptr<MshData> _msh, int no, F0Ps &_f) : Group(_msh, no), f(_f)
{
    getNodes();
}
Assign::Assign(std::shared_ptr<MshData> _msh, std::string const &name, F0Ps &_f) : Group(_msh, name), f(_f)
{
    getNodes();
}

/**
 * @brief Prescribe value to all nodes (default implementation)
 * 
 * Value can be constant or position dependent
 * @authors Adrien Crovato, Romain Boman
 */
void Assign::apply(std::vector<double> &vec)
{
    for (auto n : nodes)
        vec[n->row] = f.eval(n->pos);
}

/**
 * @brief Get all nodes belonging to boundary tag
 * @authors Adrien Crovato, Romain Boman
 */
void Assign::getNodes()
{
    for (auto e : tag->elems)
        for (auto n : e->nodes)
            nodes.push_back(n);
    std::sort(nodes.begin(), nodes.end());
    auto it = std::unique(nodes.begin(), nodes.end());
    nodes.resize(std::distance(nodes.begin(), it));
}

Initial::Initial(std::shared_ptr<tbox::MshData> _msh, int no, F0Ps &_f) : Assign(_msh, no, _f)
{
}
Initial::Initial(std::shared_ptr<tbox::MshData> _msh, std::string const &name, F0Ps &_f) : Assign(_msh, name, _f)
{
}
void Initial::write(std::ostream &out) const
{
    out << "flow::Initial condition on " << *tag << ")\n";
    for (auto n : nodes)
    {
        out << '\t' << *n << ": (val=" << f.eval(n->pos) << '\n';
    }
}

Dirichlet::Dirichlet(std::shared_ptr<tbox::MshData> _msh, int no, F0Ps &_f) : Assign(_msh, no, _f)
{
}
Dirichlet::Dirichlet(std::shared_ptr<tbox::MshData> _msh, std::string const &name, F0Ps &_f) : Assign(_msh, name, _f)
{
}
void Dirichlet::write(std::ostream &out) const
{
    out << "flow::Dirichlet boundary condition on " << *tag << ")\n";
    for (auto n : nodes)
    {
        out << '\t' << *n << ": (val=" << f.eval(n->pos) << '\n';
    }
}