#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Compute lifting (linear or nonlinear) viscous flow around a NACA 0012
# Amaury Bilocq
#
# Test the viscous-inviscid interaction scheme
# Reference to the master's thesis: http://hdl.handle.net/2268/252195
# Reference test cases with Naca0012 (different from master's thesis): 
# 1) Incompressible: Re = 1e7, M_inf = 0, alpha = 5°, p = 2, m = 7, tol = 10^-4, msTE = 0.01, msLE = 0.001
#    -> nIt = 32, Cl = 0.55, Cd = 0.0061, xtrTop = 0.0611, xtrBot = 0.7392
# 2) Compressible: Re = 1e7, M_inf = 5, alpha = 5°, p = 2, m = 7, tol = 10^-4, msTE = 0.01, msLE = 0.001
#    -> nIt = 45, Cl = 0.64, Cd = 0.0065, xtrTop = 0.0497, xtrBot = 0.7338
# 3) Separated: Re = 1e7, M_inf = 0, alpha = 12°, p = 2, m = 5, tol = 2.5*10^-3, msTE = 0.01, msLE = 0.00075
#    -> nIt = 39, Cl = 1.30 , Cd = 0.011, xtrTop = 0.010, xtrBot = 1
#
# CAUTION
# This test is provided to ensure that the solver works properly.
# Mesh refinement may have to be performed to obtain physical results.

import math
import flow.utils as floU
import flow.default as floD
import flow.viscous.solver as floVS
import flow.viscous.coupler as floVC
import tbox
import tbox.utils as tboxU
import fwk
from fwk.testing import *
from fwk.coloring import ccolors

def main():
    # timer
    tms = fwk.Timers()
    tms['total'].start()

    # define flow variables
    Re = 1e7
    alpha = 5*math.pi/180
    U_inf = [math.cos(alpha), math.sin(alpha)] # norm should be 1
    M_inf = 0.5
    M_crit = 5 # Squared critical Mach number (above which density is modified)
    c_ref = 1
    dim = 2

    # define filter parameters and tolerance of the VII 
    p = 2
    m = 5
    tol = 1e-4

    # mesh the geometry
    print(ccolors.ANSI_BLUE + 'PyMeshing...' + ccolors.ANSI_RESET)
    tms['msh'].start()
    pars = {'xLgt' : 5, 'yLgt' : 5, 'msF' : 1, 'msTe' : 0.01, 'msLe' : 0.001}
    msh, gmshWriter = floD.mesh(dim, 'models/n0012.geo', pars, ['field', 'airfoil', 'downstream'])
    tms['msh'].stop()

    # set the problem and add medium, initial/boundary conditions
    tms['pre'].start()
    pbl, _, _, bnd, blw = floD.problem(msh, dim, alpha, 0., M_inf, M_crit, c_ref, c_ref, 0.25, 0., 0., 'airfoil', te = 'te', vsc = True)
    tms['pre'].stop()

    # solve viscous problem
    print(ccolors.ANSI_BLUE + 'PySolving...' + ccolors.ANSI_RESET)
    tms['solver'].start()
    isolver = floD.newton(pbl)
    vsolver = floVS.Solver(Re, p, m)
    coupler = floVC.Coupler(isolver, vsolver, blw[0], blw[1], tol, gmshWriter)
    coupler.run()
    tms['solver'].stop()

    # extract Cp
    Cp = floU.extract(bnd.groups[0].tag.elems, isolver.Cp)
    tboxU.write(Cp, 'Cp_airfoil.dat', '%1.5e', ', ', 'x, y, z, Cp', '')
    # display results
    print(ccolors.ANSI_BLUE + 'PyRes...' + ccolors.ANSI_RESET)
    print('      Re        M    alpha       Cl       Cd      Cdp      Cdf       Cm')
    print('{0:6.1f}e6 {1:8.2f} {2:8.1f} {3:8.4f} {4:8.4f} {5:8.4f} {6:8.4f} {7:8.4f}'.format(Re/1e6, M_inf, alpha*180/math.pi, isolver.Cl, vsolver.Cd, vsolver.Cdp, vsolver.Cdf, isolver.Cm))

    # display timers
    tms['total'].stop()
    print(ccolors.ANSI_BLUE + 'PyTiming...' + ccolors.ANSI_RESET)
    print('CPU statistics')
    print(tms)

    # visualize solution and plot results
    floD.initViewer(pbl)
    tboxU.plot(Cp[:,0], Cp[:,3], 'x', 'Cp', 'Cl = {0:.{3}f}, Cd = {1:.{3}f}, Cm = {2:.{3}f}'.format(isolver.Cl, vsolver.Cd, isolver.Cm, 4), True)

    # check results
    print(ccolors.ANSI_BLUE + 'PyTesting...' + ccolors.ANSI_RESET)
    tests = CTests()
    if Re == 1e7 and M_inf == 0 and alpha == 5*math.pi/180:
        tests.add(CTest('Cl', isolver.Cl, 0.55, 5e-2)) # Xfoil 0.58
        tests.add(CTest('Cd', vsolver.Cd, 0.0062, 0.01))
        tests.add(CTest('Cdp', vsolver.Cdp, 0.0018, 0.01))
        tests.add(CTest('xtr_top', vsolver.xtr[0], 0.061, 0.05)) # Xfoil 0.056
        tests.add(CTest('xtr_bot', vsolver.xtr[1], 0.740, 0.05))
    elif Re == 1e7 and M_inf == 0.5 and alpha == 5*math.pi/180:
        tests.add(CTest('Cl', isolver.Cl, 0.65, 5e-2)) # Xfoil 0.69
        tests.add(CTest('Cd', vsolver.Cd, 0.0067, 0.01))
        tests.add(CTest('Cdp', vsolver.Cdp, 0.0025, 0.01))
        tests.add(CTest('xtr_top', vsolver.xtr[0], 0.049, 0.05)) # Xfoil 0.038
        tests.add(CTest('xtr_bot', vsolver.xtr[1], 0.736, 0.05))
    elif Re == 1e7 and M_inf == 0 and alpha == 12*math.pi/180:
        tests.add(CTest('Cl', isolver.Cl, 1.30, 5e-2)) # Xfoil 1.39
        tests.add(CTest('Cd', vsolver.Cd, 0.0011, 0.01))
        tests.add(CTest('Cdp', vsolver.Cdp, 0.0064, 0.01))
        tests.add(CTest('xtr_top', vsolver.xtr[0], 0.010, 0.05)) # Xfoil 0.008
        tests.add(CTest('xtr_bot', vsolver.xtr[1], 1.000, 0.05))
    else:
        raise Exception('Test not defined for this flow')
    
    tests.run()

    # eof
    print('')

if __name__ == "__main__":
    main()
