#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Compute the (non)linear flow around cylinder, with various B.C.
# Adrien Crovato
#
# Test solver convergence and boundary conditions
#
# CAUTION
# This test is provided to ensure that the solver works properly.
# Mesh refinement may have to be performed to obtain physical results.

import math
import flow.default as floD
import fwk
from fwk.testing import *
from fwk.coloring import ccolors 

def main():
    # timer
    tms = fwk.Timers()
    tms['total'].start()

    # define flow variables
    alpha = 3*math.pi/180
    U_inf = [math.cos(alpha), math.sin(alpha)] # norm should be 1
    M_inf = 0.1
    M_crit = 5 # Squared critical Mach number (above which density is modified)
    dim = 2

    # define dimension and mesh size
    dmt = 1.0 # cylinder diameter
    lgt = 8.0 # channel length
    hgt = 6.0 # channel height
    l_ref = dmt # reference length
    fms = 0.5 # farfield mesh size
    nms = 0.1 # nearfield mesh size

    # mesh the geometry
    print(ccolors.ANSI_BLUE + 'PyMeshing...' + ccolors.ANSI_RESET)
    tms['msh'].start()
    pars = {'dmt' : dmt, 'lgt' : lgt, 'hgt' : hgt, 'msN' : nms, 'msF' : fms}
    msh, gmshWriter = floD.mesh(dim, 'models/cylinder.geo', pars, ['field', 'cylinder', 'downstream'])
    tms['msh'].stop()

    # set the problem and add medium, initial/boundary conditions
    tms['pre'].start()
    pbl, dirichlet, wake, bnd, _ = floD.problem(msh, dim, alpha, 0., M_inf, M_crit, l_ref, l_ref, 0., 0., 0., 'cylinder', te = 'te')
    tms['pre'].stop()

    # solve problem
    print(ccolors.ANSI_BLUE + 'PySolving...' + ccolors.ANSI_RESET)
    tms['picard'].start()
    solver0 = floD.picard(pbl)
    cnvrgd0 = solver0.run()
    solver0.save(gmshWriter)
    tms['picard'].stop()
    tms['newton'].start()
    solver1 = floD.newton(pbl)
    cnvrgd1 = solver1.run()
    solver1.save(gmshWriter)
    tms['newton'].stop()

    # compute mean error in Boundary, Neumann boundary condition (works only for Line2 elements)
    gradn0 = []
    gradn1 = []
    for e in bnd.groups[0].tag.elems:
        gradx0 = 0
        grady0 = 0
        gradx1 = 0
        grady1 = 0
        nx = e.nodes[0].pos[1] - e.nodes[1].pos[1]
        ny = e.nodes[1].pos[0] - e.nodes[0].pos[0]
        for n in e.nodes:
            gradx0 += solver0.U[n.row][0] / e.nodes.size()
            grady0 += solver0.U[n.row][1] / e.nodes.size()
            gradx1 += solver1.U[n.row][0] / e.nodes.size()
            grady1 += solver1.U[n.row][1] / e.nodes.size()
        gradn0.append(gradx0 * nx + grady0 * ny)
        gradn1.append(gradx1 * nx + grady1 * ny)
    errNeu0 = sum(gradn0) / len(gradn0)
    errNeu1 = sum(gradn1) / len(gradn1)
    # compute largest error in Dirichlet boundary condition
    errDir0 = 0.
    errDir1 = 0.
    for n in dirichlet.nodes:
        errCur = solver0.phi[n.row] - (n.pos[0]*U_inf[0] + n.pos[1]*U_inf[1])
        if (abs(errCur) > errDir0):
            errDir0 = abs(errCur)
        errCur = solver1.phi[n.row] - (n.pos[0]*U_inf[0] + n.pos[1]*U_inf[1])
        if (abs(errCur) > errDir1):
            errDir1 = abs(errCur)
    # compute largest error in Wake boundary condition
    errWak0 = 0.
    errWak1 = 0.
    for n in wake.nodMap:
        errCur = solver0.Cp[n.row] - solver0.Cp[wake.nodMap[n].row]
        if (abs(errCur) > errWak0):
            errWak0 = abs(errCur)
        errCur = solver1.Cp[n.row] - solver1.Cp[wake.nodMap[n].row]
        if (abs(errCur) > errWak1):
            errWak1 = abs(errCur)

    # display timers
    tms['total'].stop()
    print(ccolors.ANSI_BLUE + 'PyTiming...' + ccolors.ANSI_RESET)
    print('CPU statistics')
    print(tms)

    # visualize solution
    floD.initViewer(pbl)

    # check results
    print(ccolors.ANSI_BLUE + 'PyTesting...' + ccolors.ANSI_RESET)
    if not cnvrgd0 or not cnvrgd1:
        raise Exception(ccolors.ANSI_RED + 'Flow solver failed to converge!' + ccolors.ANSI_RESET)
    tests = CTests()
    tests.add(CTest('Picard: iteration count', solver0.nIt, 12, 1, forceabs=True))
    tests.add(CTest('Picard: Neumann B.C. mean error', errNeu0, 0., 1e-2))
    tests.add(CTest('Picard: Dirichlet B.C. max error', errDir0, 0., 1e-8))
    tests.add(CTest('Picard: Wake/Kutta B.C. max error', errWak0, 0., 1e-2))
    tests.add(CTest('Newton: iteration count', solver1.nIt, 3, 1, forceabs=True))
    tests.add(CTest('Newton: Neumann B.C. mean error', errNeu1, 0., 1e-2))
    tests.add(CTest('Newton: Dirichlet B.C. max error', errDir1, 0., 1e-8))
    tests.add(CTest('Newton: Wake/Kutta B.C. max error', errWak1, 0., 1e-2))
    tests.run()

    # eof
    print('')

if __name__ == "__main__":
    main()
