#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Compute the flow around a NACA 0012 on a deformed mesh
# Adrien Crovato
#
# Test the mesh deformation process
#
# CAUTION
# This test is provided to ensure that the solver works properly.
# Mesh refinement may have to be performed to obtain physical results.

import numpy as np
import flow.utils as floU
import flow.default as floD
import tbox.utils as tboxU
import fwk
from fwk.testing import *
from fwk.coloring import ccolors 

def main():
    # timer
    tms = fwk.Timers()
    tms['total'].start()

    # define flow variables
    alpha0 = 0 # must be zero for this testfile
    M_inf = 0.3
    M_crit = 5 # Squared critical Mach number (above which density is modified)
    c_ref = 1
    dim = 2
    # parameters for mesh deformation
    alfa = 10*np.pi/180
    xc = 0.25
    dx = 0.2
    dy = 0.5

    # mesh the geometry
    print(ccolors.ANSI_BLUE + 'PyMeshing...' + ccolors.ANSI_RESET)
    tms['msh'].start()
    pars = {'xLgt' : 5, 'yLgt' : 5, 'msF' : 1., 'msTe' : 0.01, 'msLe' : 0.01}
    msh, gmshWriter = floD.mesh(dim, 'models/n0012.geo', pars, ['field', 'airfoil', 'downstream'])
    # create mesh deformation handler
    mshDef = floD.meshMorpher(msh, dim, ['airfoil'])
    # mesh the reference geometry
    pars = {'xLgt' : 5, 'yLgt' : 5, 'msF' : 1., 'msTe' : 0.01, 'msLe' : 0.01, 'angle' : alfa, 'xRot' : xc}
    msh_ref, gmshWriter_ref = floD.mesh(dim, 'models/n0012.geo', pars, ['field', 'airfoil', 'downstream'])
    tms['msh'].stop()

    # set the problem and add medium, initial/boundary conditions
    tms['pre'].start()
    pbl, _, _, bnd, _ = floD.problem(msh, dim, alpha0, 0., M_inf, M_crit, c_ref, c_ref, xc, 0., 0., 'airfoil', te = 'te')
    # set the refrence problem and add medium, initial/boundary conditions
    pbl_ref, _, _, bnd_ref, _ = floD.problem(msh_ref, dim, alpha0, 0., M_inf, M_crit, c_ref, c_ref, xc, 0., 0., 'airfoil', te = 'te')
    tms['pre'].stop()

    # solver problem for 0°AOA
    print(ccolors.ANSI_BLUE + 'PySolving...' + ccolors.ANSI_RESET)
    tms['solver0'].start()
    solver = floD.newton(pbl)
    solver.run()
    tms['solver0'].stop()

    # deform the mesh (dummy rotation + translation)
    print(ccolors.ANSI_BLUE + 'PyDeforming...' + ccolors.ANSI_RESET)
    tms['deform'].start()
    mshDef.savePos()
    rMat = np.matrix([[np.cos(alfa), np.sin(alfa)], [-np.sin(alfa), np.cos(alfa)]])
    vNods = {}
    for e in bnd.groups[0].tag.elems:
        for n in e.nodes:
            if n.no not in vNods:
                vNods[n.no] = n
    for no in vNods:
        nod = np.matrix([[vNods[no].pos[0]-xc], [vNods[no].pos[1]]])
        new = rMat * nod
        vNods[no].pos[0] = new[0,0] + dx + xc
        vNods[no].pos[1] = new[1,0] + dy
    # deform the mesh
    mshDef.deform()
    gmshWriter.save("n0012_def")
    tms['deform'].stop()

    # solve problem for alfa AOA
    print(ccolors.ANSI_BLUE + 'PySolving...' + ccolors.ANSI_RESET)
    tms['solver1'].start()
    solver.run()
    solver.save(gmshWriter)
    tms['solver1'].stop()

    # compute reference solution
    print(ccolors.ANSI_BLUE + 'PySolving...' + ccolors.ANSI_RESET)
    tms['solver_ref'].start()
    solver_ref = floD.newton(pbl_ref)
    solver_ref.run()
    solver_ref.save(gmshWriter_ref)
    tms['solver_ref'].stop()

    # extract Cp
    Cp = floU.extract(bnd.groups[0].tag.elems, solver.Cp)
    tboxU.write(Cp, 'Cp_airfoil.dat', '%1.5e', ', ', 'x, y, z, Cp', '')
    Cp_ref = floU.extract(bnd_ref.groups[0].tag.elems, solver_ref.Cp)
    tboxU.write(Cp, 'Cp_airfoil_ref.dat', '%1.5e', ', ', 'x, y, z, Cp', '')
    # display results
    print(ccolors.ANSI_BLUE + 'PyRes...' + ccolors.ANSI_RESET)
    print('  case        M    alpha       Cl       Cd       Cm')
    print('  true {0:8.2f} {1:8.1f} {2:8.4f} {3:8.4f} {4:8.4f}'.format(M_inf, alfa*180/np.pi, solver.Cl, solver.Cd, solver.Cm))
    print('   ref {0:8.2f} {1:8.1f} {2:8.4f} {3:8.4f} {4:8.4f}'.format(M_inf, alfa*180/np.pi, solver_ref.Cl, solver_ref.Cd, solver_ref.Cm))
    print('Cm will differ unless dx and dy are set to 0...')

    # display timers
    tms["total"].stop()
    print(ccolors.ANSI_BLUE + 'PyTiming...' + ccolors.ANSI_RESET)
    print("CPU statistics")
    print(tms)

    # visualize solution and plot results
    tboxU.plot(Cp[:,0], Cp[:,3], 'x', 'Cp', 'Cl = {0:.{3}f}, Cd = {1:.{3}f}, Cm = {2:.{3}f}'.format(solver.Cl, solver.Cd, solver.Cm, 4), True)
    tboxU.plot(Cp_ref[:,0], Cp_ref[:,3], 'x', 'Cp', 'Cl = {0:.{3}f}, Cd = {1:.{3}f}, Cm = {2:.{3}f}'.format(solver_ref.Cl, solver_ref.Cd, solver_ref.Cm, 4), True)

    # check results
    print(ccolors.ANSI_BLUE + 'PyTesting...' + ccolors.ANSI_RESET)
    tests = CTests()
    tests.add(CTest('solver: iteration count', solver.nIt, 5, 3, forceabs=True)) # todo: tolerance should be reset to 1 as soon as gmsh4 is used
    tests.add(CTest('solver_ref: iteration count', solver_ref.nIt, 5, 1, forceabs=True))
    tests.add(CTest('min(Cp)', min(Cp[:,3]), min(Cp_ref[:,3]), 5e-2))
    tests.add(CTest('Cl', solver.Cl, solver_ref.Cl, 5e-2))
    tests.run()

    # eof
    print('')

if __name__ == "__main__":
    main()
