#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Newton raphson method coupled with linear solver
# Amaury Bilocq

import numpy as np
from fwk.coloring import ccolors

def newton(f, x, maxIt, tol=1.0e-9):
    '''Newton procedure to solve the non linear set of  boundary layer equations.
    Boundary layer equations are parabolic in x -> downstream marching is used
    An implicit marching model is applied (Crank Nicolson) -> matrix inversion is performed
    '''
    # Compute Jacobian 
    def jacobian(f, x):
        dx = 1.0e-8
        n = len(x)
        jac = np.zeros((n,n))
        f0 = f(x) 
        for j in range(n):
            Dxj = (abs(x[j])*dx if x[j] != 0 else dx)
            x_plus = [(xi if k != j else xi+Dxj) for k, xi in enumerate(x)]
            jac[:,j] = (f(x_plus)-f0)/Dxj # numerical solution of the jacobian 
        return jac, f0
    # Solve
    for i in range(maxIt):
        jac, f0 = jacobian(f,x)
        if np.sqrt(np.dot(f0,f0)/len(x)) < tol and x[0]>0 and  x[1]>0: return x 
        dx = np.linalg.solve(jac, -f0)
        x = x + dx
        x[1] = max(x[1],1.0005) # to avoid too low shape parameter. Solver must be redone to be more robust
        if np.sqrt(np.dot(dx, dx)) < tol*max(abs(any(x)),1) and x[0]>0 and x[1]>0: return x
    # Error
    raise RuntimeError(ccolors.ANSI_RED + 'Viscous Newton solver - too many iterations, aborting!\n' + ccolors.ANSI_RESET)
