#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""@package docstring
This test runs the four different cases as defined in table 2 of the paper
with a SIMPLE-GS-GS preconditioner.
"""

import katoptron.CMAME.model as model

import katoptron as m
import tbox
import tbox.gmsh as gmsh
import shlex
import subprocess
import os
import numpy as np

from fwk.wutils import parseargs

from PyTrilinos import Teuchos

import katoptron.utilities as u


def get_parameters(case_id, first_sample_id):
    """
    This function returns the python dictionary that includes 
    the parameters for a given test case
    """
    p = model.defaut_params

    if case_id == 1:
        nuw = 128
        p['Correlation length'] = 20.
        p['Dispersion level'] = 0.1
    elif case_id == 2:
        nuw = 128
        p['Correlation length'] = 20.
        p['Dispersion level'] = 0.3
    elif case_id == 3:
        nuw = 128
        p['Correlation length'] = 5.
        p['Dispersion level'] = 0.1
    elif case_id == 4:
        nuw = 128
        p['Correlation length'] = 5.
        p['Dispersion level'] = 0.3
    elif case_id == 5:
        nuw = 128
        p['Correlation length'] = 2.
        p['Dispersion level'] = 0.1
    elif case_id == 6:
        nuw = 128
        p['Correlation length'] = 2.
        p['Dispersion level'] = 0.3

    p['Wavenumber cutoff'] = 8.
    p['Wavenumber discretization points'] = nuw
    p['Number Z of previously drawn samples'] = first_sample_id * nuw**2
    p['Number Phi of previously drawn samples'] = first_sample_id * nuw**2

    p['Is exp'] = False
    return p


def evaluate_one_ensemble(ensemble_size, case_id, first_sample_id, msh, comm,
                          file_dir):
    """
    This function evaluates one ensemble for a given test case
    """
    work_dir = os.getcwd()
    args = parseargs()

    p = get_parameters(case_id, first_sample_id)

    pbl = m.Problem(msh, comm)

    # Units are in mm
    E = 70000 * np.ones(ensemble_size)  # 70e9 Pa = 70000 MPa
    nu = 0.35 * np.ones(ensemble_size)
    k = 0.
    d = 0.
    m.Medium(pbl, "body", "test", E, nu, k, d)

    m.Dirichlet(pbl, "clamped xyz", "Clamped", 1, 0., 1, 0., 1, 0., 0, 0.,
                ensemble_size)
    m.Dirichlet(pbl, "clamped xy", "Clamped", 1, 0., 1, 0., 0, 0., 0, 0.,
                ensemble_size)
    m.Dirichlet(pbl, "clamped y", "Clamped", 0, 0., 1, 0., 0, 0., 0, 0.,
                ensemble_size)

    f = 700 * np.ones(ensemble_size)  # 700e6 Pa = 700 MPa
    zero = np.zeros(ensemble_size)
    one = np.ones(ensemble_size)
    m.Neumann(pbl, 'load 1', 'load', 0, zero, 1, f, 0, zero, 0, one,
              ensemble_size)
    m.Neumann(pbl, 'load 2', 'load', 0, zero, 1, -f, 0, zero, 0, one,
              ensemble_size)

    solverList = Teuchos.ParameterList()
    solverList['Ensemble Convergence Tolerance'] = 1e-6
    solverList['Maximum Iterations'] = 1000
    solverList['Verbosity'] = 33
    solverList['Flexible Gmres'] = False
    solverList['Output Style'] = 1
    solverList['Output Frequency'] = 1

    prec = 3

    if prec == 1:
        mueluParams = Teuchos.ParameterList()
        mueluParams['number of equations'] = 3
        mueluParams['max levels'] = 1
        mueluParams['coarse: type'] = "CHEBYSHEV"

        mueluParamsSub2 = Teuchos.ParameterList()
        mueluParamsSub2['chebyshev: degree'] = 20

        mueluParams['verbosity'] = "high"

        mueluParams['coarse: params'] = mueluParamsSub2
    elif prec == 2:
        mueluParams = Teuchos.ParameterList()
        mueluParams['number of equations'] = 3
        mueluParams['max levels'] = 1
        mueluParams['coarse: type'] = "RELAXATION"

        mueluParamsSub2 = Teuchos.ParameterList()
        mueluParamsSub2['relaxation: type'] = "MT Gauss-Seidel"
        mueluParamsSub2['relaxation: sweeps'] = 10
        mueluParamsSub2['relaxation: damping factor'] = 1.75

        mueluParams['verbosity'] = "high"

        mueluParams['coarse: params'] = mueluParamsSub2
    elif prec == 3:
        mueluParams = Teuchos.ParameterList()
        mueluParams['number of equations'] = 3
        mueluParams['max levels'] = 2
        mueluParams['smoother: type'] = "RELAXATION"

        mueluParamsSub2 = Teuchos.ParameterList()
        mueluParamsSub2['relaxation: type'] = "MT Gauss-Seidel"
        mueluParamsSub2['relaxation: sweeps'] = 10
        mueluParamsSub2['relaxation: damping factor'] = 1.75

        mueluParams['verbosity'] = "high"

        mueluParams['smoother: params'] = mueluParamsSub2

        mueluParams['coarse: type'] = "Klu"
        mueluParams['coarse: max size'] = 18

    solverList['convert MueLu xml file'] = True
    solverList['mueluParams'] = mueluParams

    solverList['Print Teuchos timers'] = True

    solverList['gamma'] = p['gamma']
    solverList['Use preconditioner'] = True

    solverList['type'] = 'BlockGmres'
    solverList['Maximum active set iteration'] = 1

    solverList['Write matrix and vectors'] = p['Write matrix and vectors']
    solverList['Write txt files'] = p['Write txt files']

    if p["Is exp"]:
        num_random_variables = randomVariable.shape[1]
    else:
        num_random_variables = 0

    for i in range(0, num_random_variables):
        rand = randomVariable[:, i]
        pbl.add(m.RandomVariable(rand))

    randomParams = Teuchos.ParameterList()

    randomParams['Is random'] = True
    randomParams['Is exp'] = p['Is exp']
    if p["Is exp"]:
        randomParams['Begin X'] = 0.
        randomParams['Begin Y'] = 0.
        randomParams['Begin Z'] = 0.
        randomParams['End X'] = 10.
        randomParams['End Y'] = 10.
        randomParams['End Z'] = 1.
        randomParams['Number random variables'] = num_random_variables
        randomParams['Correlation length X'] = p['Correlation length X']
        randomParams['Correlation length Y'] = p['Correlation length Y']
        randomParams['Correlation length Z'] = p['Correlation length Z']
        randomParams['Number of KL Terms X'] = p['Number of KL Terms X']
        randomParams['Number of KL Terms Y'] = p['Number of KL Terms Y']
        randomParams['Number of KL Terms Z'] = p['Number of KL Terms Z']
        randomParams['Maximum Nonlinear Solver Iterations'] = 5000
        randomParams['Bound Perturbation Size'] = 1E-10
    else:
        randomParams['Correlation length'] = p['Correlation length']
        randomParams['Wavenumber cutoff'] = p['Wavenumber cutoff']
        randomParams['Wavenumber discretization points'] = p[
            'Wavenumber discretization points']
        randomParams['Number Z of previously drawn samples'] = p[
            'Number Z of previously drawn samples']
        randomParams['Number Phi of previously drawn samples'] = p[
            'Number Phi of previously drawn samples']

    randomParams['Mean'] = E[0] / (2. * (1. + nu[0]))
    randomParams['Dispersion level'] = p['Dispersion level']

    solverList['randomParams'] = randomParams

    slv = m.IterativeSolver(pbl, args.k, solverList, 3, ensemble_size)
    slv.start()
    f = open('timers.txt', 'w')
    print(slv.getTimers(), file=f)
    f.closed


def evaluate_all_ensembles(ensemble_sizes, case_ids, N, msh, comm, file_dir):
    """
    This function loop over the test cases, the ensemble sizes,
    and the ensembles
    """
    for case_id in case_ids:
        directory_case_id = 'case_' + str(case_id)
        if not os.path.exists(directory_case_id):
            os.makedirs(directory_case_id)
        os.chdir(directory_case_id)
        for ensemble_size in ensemble_sizes:
            directory = str(ensemble_size)
            if not os.path.exists(directory):
                os.makedirs(directory)
            os.chdir(directory)

            N_ensemble = N // ensemble_size
            for ensemble_index in range(0, N_ensemble):
                directory = str(ensemble_index)
                if not os.path.exists(directory):
                    os.makedirs(directory)
                os.chdir(directory)

                first_sample_id = ensemble_index * ensemble_size

                evaluate_one_ensemble(ensemble_size, case_id, first_sample_id,
                                      msh, comm, file_dir)
                os.chdir('..')
            os.chdir('..')
        os.chdir('..')


def main():
    """
    This loads the mesh and define the case ids and ensemble sizes
    """
    comm, rank, size = m.utilities.import_MPI()

    args = parseargs()

    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()

    # Units in the meshes are in mm
    msh = gmsh.MeshLoader('plate.geo', file_dir).execute()

    N = 32
    ensemble_sizes = [1, 8, 16, 32]
    case_ids = [1, 2, 3, 4, 5, 6]

    evaluate_all_ensembles(ensemble_sizes, case_ids, N, msh, comm, file_dir)


if __name__ == "__main__":
    main()
