import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import axes3d
import matplotlib.tri as mtri
from matplotlib import cm

import os

from katoptron.Juelich.UQ.kde import *
import katoptron.utilities as u

import scipy as sc


def build_exponents_L2(order):
    exponents = np.zeros(((order + 1)**2, 2), dtype=int)
    exponents[:, 0], exponents[:, 1] = np.divmod(np.arange((order + 1)**2),
                                                 order + 1)
    return exponents


def build_exponents_L1(order):
    exponents = build_exponents_L2(order)
    tmp = exponents[np.where((exponents.T[0] + exponents.T[1]) <= order), :]
    return tmp[0, :, :]


def build_PCE_system(x,
                     y,
                     weights,
                     exponents,
                     p0=np.array([0, 0]),
                     scale=np.array([1, 1]),
                     monomial=True):
    n = len(x)
    m = exponents.shape[0]
    Z = np.zeros((n, m))
    W = np.zeros((n, n))

    if not monomial:
        u = 1. - (x - 30.) / 76.
        v = -1. + (y - 30.) / 76.

    for i in range(0, m):
        if monomial:
            Z[:, i] = (((x - p0[0]) / scale[0]))**exponents[i, 0] * (
                ((y - p0[1]) / scale[1]))**exponents[i, 1]
        else:
            p_m_x = np.ones(u.shape)
            index = np.argwhere(v < 1)
            p_m_x[index] = (
                (2 * u[index] + v[index] + 1) / (1 - v[index]))
            p_m = sc.special.eval_legendre(exponents[i, 0], p_m_x)
            p_n = sc.special.eval_jacobi(exponents[i, 1],
                                         2 * exponents[i, 0] + 1, 0, v)
            Z[:, i] = p_m * (((1 - v) / 2))**exponents[i, 0] * p_n
    for i in range(0, n):
        W[i, i] = weights[i]

    ZTW = Z.T.dot(W)
    ZTWZ = ZTW.dot(Z)

    ZTWZinv = np.linalg.pinv(ZTWZ)
    Ainv = ZTWZinv.dot(ZTW)
    return Ainv


def evaluate_PCE(x,
                 y,
                 coefficients,
                 exponents,
                 p0=np.array([0, 0]),
                 scale=np.array([1, 1]),
                 monomial=True):
    n = len(x)
    m = exponents.shape[0]
    z = np.zeros((n, ))

    if not monomial:
        u = 1. - (x - 30.) / 76.
        v = -1. + (y - 30.) / 76.

    for i in range(0, m):
        if monomial:
            z = z + coefficients[i] * (
                ((x - p0[0]) / scale[0]))**exponents[i, 0] * (
                    ((y - p0[1]) / scale[1]))**exponents[i, 1]
        else:
            p_m_x = np.ones(u.shape)
            index = np.argwhere(v < 1)
            p_m_x[index] = (
                (2 * u[index] + v[index] + 1) / (1 - v[index]))
            p_m = sc.special.eval_legendre(exponents[i, 0], p_m_x)
            p_n = sc.special.eval_jacobi(exponents[i, 1],
                                         2 * exponents[i, 0] + 1, 0, v)
            z = z + coefficients[i] * p_m * (
                ((1 - v) / 2))**exponents[i, 0] * p_n
    return z


def triangular_mesh(n_x, n_y, x_min=30, x_max=180, y_min=30, y_max=180):
    x = np.linspace(x_min, x_max, n_x, endpoint=True)
    y = np.linspace(y_min, y_max, n_y, endpoint=True)
    x, y = np.meshgrid(x, y)
    x = x.flatten()
    y = y.flatten()

    x_TRI = x[np.where(x >= y)]
    y_TRI = y[np.where(x >= y)]

    tri = mtri.Triangulation(x_TRI, y_TRI)

    return x_TRI, y_TRI, tri


def compute_coeff(Ainv, z):
    coefficients = Ainv.dot(z)
    return coefficients


def write_patch(tri, z, filename):
    f = open(filename + "_data.txt", "w+")
    print("x y c", file=f)
    for i in range(0, tri.triangles.shape[0]):
        for j in range(0, 3):
            index = tri.triangles[i, j]
            print(str(tri.x[index]) + " " + str(
                tri.y[index]) + " " + str(z[index]), file=f)
        print(" ", file=f)

    f.close()
    f = open(filename + "_range.txt", "w+")
    print(np.amin(z), file=f)
    print(np.amax(z), file=f)
    f.close()


def main():
    file_dir = os.path.dirname(__file__)

    N = 20
    loaded = np.load(file_dir + '/../samples_files/Xiao_' + str(N) + '.npz')
    samples = loaded['samples']
    N = samples.shape[0]

    x = samples[0:N, 0]
    y = samples[0:N, 1]
    weights = loaded['weights']

    #qois = np.load('../katoptron_Juelich_UQ_mirror/qois_xiao.data.npy')
    qois = np.load('../katoptron_Juelich_UQ_mirror/qois_xiao_2.data.npy')
    qois = np.load('../katoptron_Juelich_UQ_mirror/A2_xiao_20.data.npy')
    qois = np.load('../katoptron_Juelich_UQ_mirror/B2_xiao_20.data.npy')

    preloads_A2_N = np.loadtxt(
        '../katoptron_Juelich_UQ_preload/preloads_A2_N.txt')
    preloads_B2_N = np.loadtxt(
        '../katoptron_Juelich_UQ_preload/preloads_B2_N.txt')

    x_0 = [30, 30]
    scale = [150, 150]

    n_x = 30
    n_y = 30

    monomial = False

    order = 10
    exponents = build_exponents_L1(order)
    Ainv = build_PCE_system(x,
                            y,
                            weights,
                            exponents,
                            x_0,
                            scale,
                            monomial=monomial)

    for z_index in range(0, 9):
        z = qois[z_index, :]
        filename = 'qoi_' + str(z_index)

        coefficients = compute_coeff(Ainv, z)
        print(len(coefficients))

        x_PCE, y_PCE, tri = triangular_mesh(n_x, n_y)

        z_PCE = evaluate_PCE(x_PCE,
                             y_PCE,
                             coefficients,
                             exponents,
                             x_0,
                             scale,
                             monomial=monomial)

        write_patch(tri, z_PCE, filename)

    for z_index in range(0, 3):
        z = preloads_A2_N[:, z_index]
        filename = 'qoi_preload_A_' + str(z_index)

        coefficients = compute_coeff(Ainv, z)
        print(len(coefficients))

        x_PCE, y_PCE, tri = triangular_mesh(n_x, n_y)

        z_PCE = evaluate_PCE(x_PCE,
                             y_PCE,
                             coefficients,
                             exponents,
                             x_0,
                             scale,
                             monomial=monomial)

        write_patch(tri, -0.001 * z_PCE, filename)
        z = preloads_B2_N[:, z_index]
        filename = 'qoi_preload_B_' + str(z_index)

        coefficients = compute_coeff(Ainv, z)
        print(len(coefficients))

        x_PCE, y_PCE, tri = triangular_mesh(n_x, n_y)

        z_PCE = evaluate_PCE(x_PCE,
                             y_PCE,
                             coefficients,
                             exponents,
                             x_0,
                             scale,
                             monomial=monomial)

        write_patch(tri, -0.001 * z_PCE, filename)


if __name__ == "__main__":
    main()
