#include <Xpetra_MatrixFactory.hpp>
#include <Xpetra_Matrix.hpp>
#include <Xpetra_MatrixMatrix.hpp>
#include <Xpetra_CrsMatrixWrap.hpp>
#include <Xpetra_BlockedCrsMatrix.hpp>
#include <Xpetra_CrsMatrix.hpp>
#include <MueLu_Utilities_decl.hpp>

namespace katoptron
{

/**
 * @brief DiscreteProblem constructor
 * 
 * This constructor constructs a katoptron::Map, a katoptron::Domain, a katoptron::Loads, and a katoptron::Algebraic.
 * The last three objects are stored as member data of the katoptron::DiscreteProblem object and include all
 * the information related to the problem to solve. See the documentation of the member data for more information on
 * the purpose of each class.
 * 
 * Arguments:
 * <ul>
 *  <li> A problem object,
 *  <li> An RCP to a ParameterList object which describes the random fields,
 *  <li> A Kokkos View of the values used in the discretization of the random fields.
 * </ul>
 */
template <typename scalar>
DiscreteProblem<scalar>::DiscreteProblem(Problem &pbl, size_t _numPrimalDPN,
                                         Teuchos::RCP<Teuchos::ParameterList> randomParams,
                                         Kokkos::View<scalar *, Kokkos::LayoutLeft> m_rv_values)
{
    numPrimalDPN = _numPrimalDPN;

    LinearSolver::getTimers()["map"].start();
    Teuchos::RCP<Map> map = Teuchos::rcp(new Map(pbl, numPrimalDPN));
    LinearSolver::getTimers()["map"].stop();
    LinearSolver::getTimers()["domain"].start();
    domain = Teuchos::rcp(new Domain<scalar>(pbl, map, randomParams, m_rv_values));
    LinearSolver::getTimers()["domain"].stop();
    LinearSolver::getTimers()["loads"].start();
    loads = Teuchos::rcp(new Loads<scalar>(pbl, map, domain));
    LinearSolver::getTimers()["loads"].stop();
    algebraic = Teuchos::rcp(new Algebraic<scalar>(map, domain));
}

/**
 * @brief Update the activity of the Lagrange multipliers
 * and return a Mask which stores true for the converged samples.
 * 
 * Argument:
 * <ul>
 *  <li> active_set_iteration: ID of the current active set iteration.
 * </ul>
 */
template <typename scalar>
Mask<scalar> DiscreteProblem<scalar>::updateActivity(int active_set_iteration)
{
    using Teuchos::Array;
    using Teuchos::RCP;
    using Teuchos::rcp;

    const size_t myRank = algebraic->map->mapDofs->getComm()->getRank();

    // First, get the displacement solution,

    RCP<tpetra_mvector_type> mx =
        (Teuchos::rcp_dynamic_cast<xpetra_tmvector_type>(algebraic->vectors->solutionMultiVector))->getTpetra_MultiVector();
    RCP<tpetra_vector_type> xl = mx->getVectorNonConst(0);

    RCP<tpetra_vector_type> x = rcp(new tpetra_vector_type(algebraic->map->mapDofs, false));

    RCP<xpetra_bmvector_type> blockedSol =
        rcp(new xpetra_bmvector_type(algebraic->map->blockedMap, algebraic->vectors->solutionMultiVector));

    RCP<tpetra_vector_type> l =
        (Teuchos::rcp_dynamic_cast<xpetra_tmvector_type>(blockedSol->getMultiVector(1, true)))->getTpetra_MultiVector()->getVectorNonConst(0);

    RCP<tpetra_vector_type> gap = rcp(new tpetra_vector_type(*algebraic->vectors->initialGap, Teuchos::Copy));

    RCP<export_type> exportx = rcp(new export_type(algebraic->map->fullmap, algebraic->map->mapDofs));

    x->doExport(*xl, *exportx, Tpetra::INSERT);

    // Multiply it with B_G and remove initial gap

    // gap = 1 * B_G * x - 1 * gap

    algebraic->matrices->B_G->apply(*x, *gap, Teuchos::NO_TRANS, ((scalar)1.), ((scalar)-1.));

    // sum lambda + c * the previously computed vector
    double c = 1.;
    l->update((scalar)c, *gap, (scalar)1.); // l = l + c*gap

    // loop on the local lagrange and create the new activity

    l->template sync<Kokkos::HostSpace>();
    auto l_2d = l->template getLocalView<Kokkos::HostSpace>();

    Mask<scalar> hasConverged;
    const size_t ensemble_size = hasConverged.getSize();

    for (size_t l = 0; l < ensemble_size; ++l)
        hasConverged.set(l, true);

    bool hasConverged_old = false;
    if (active_set_iteration > 1)
        hasConverged_old = true;

    algebraic->matrices->B_2->resumeFill();
    algebraic->matrices->C->resumeFill();
    algebraic->matrices->Cb->resumeFill();

    const size_t numDPN = algebraic->map->numPrimalDPN;

    size_t numLMPN;
    if (domain->contactsList->hasAtLeastOneSticking() && numDPN == 4)
        numLMPN = 4;
    else if (domain->contactsList->hasAtLeastOneSticking())
        numLMPN = 3;
    else if (numDPN == 4)
        numLMPN = 2;
    else
        numLMPN = 1;

    const int numMyLagrangeNum = algebraic->map->mapLagrangeDofs->getNodeNumElements();

    const int numMyMechLagrangeNum = numMyLagrangeNum / numLMPN;

    for (auto local_index = 0; local_index < numMyMechLagrangeNum; ++local_index)
    {
        global_ordinal_type lagrange_id = algebraic->map->mapLagrangeDofs->getGlobalElement(local_index);
        global_ordinal_type node_id = algebraic->map->lm_to_dof[lagrange_id];
        lagrange_id = lagrange_id * numLMPN;
        local_ordinal_type interface_id = domain->contactsList->getInterfaceOfSlaveNode(node_id);

        if (domain->contactsList->isNotUpdated(interface_id))
            continue;

        scalar old_mask = old_activity(local_index);
        scalar old_old_mask = old_old_activity(local_index);
        Mask<scalar> current_mask = (l_2d(local_index * numLMPN, 0) >= 0.);

        scalar oneMask, oneNotMask;
        mask_assign<scalar>(current_mask, oneMask) = {1., 0.};
        mask_assign<scalar>(current_mask, oneNotMask) = {0., 1.};

        if (active_set_iteration > 1)
            if (!MaskLogic::AND(oneMask == old_old_mask))
                hasConverged_old = false;

        hasConverged = hasConverged && (oneMask == old_mask);

        if (!MaskLogic::AND(oneMask == old_mask))
        {
            global_ordinal_type lagrange_id = algebraic->map->mapLagrangeDofs->getGlobalElement(local_index);

            lagrange_id = lagrange_id * numLMPN;

            // If the current Lagrange multiplier is active for every sample of the current ensemble,
            // we can do the update without using the mask entries.
            if (MaskLogic::AND(current_mask))
            {
                Teuchos::ArrayView<const local_ordinal_type> local_indices;
                Teuchos::ArrayView<const scalar> values;

                if (domain->contactsList->isSticking(interface_id))
                {
                    for (size_t j = 0; j < numLMPN; ++j)
                    {
                        algebraic->matrices->B->getLocalRowView(local_index * numLMPN + j, local_indices, values);

                        for (auto k = 0; k < local_indices.size(); ++k)
                        {
                            global_ordinal_type column_index = algebraic->matrices->B->getColMap()->getGlobalElement(local_indices[k]);

                            algebraic->matrices->B_2->replaceGlobalValues(lagrange_id + j,
                                                                          Teuchos::tuple<global_ordinal_type>(column_index),
                                                                          Teuchos::tuple<scalar>(values[k]));
                        }

                        algebraic->matrices->C->replaceGlobalValues(lagrange_id + j,
                                                                    Teuchos::tuple<global_ordinal_type>(lagrange_id + j),
                                                                    Teuchos::tuple<scalar>((scalar)0.));

                        algebraic->matrices->Cb->replaceGlobalValues(lagrange_id + j,
                                                                     Teuchos::tuple<global_ordinal_type>(lagrange_id + j),
                                                                     Teuchos::tuple<scalar>((scalar)1.));
                    }
                }
                else
                {
                    algebraic->matrices->B->getLocalRowView(local_index * numLMPN, local_indices, values);

                    for (auto k = 0; k < local_indices.size(); ++k)
                    {
                        global_ordinal_type column_index = algebraic->matrices->B->getColMap()->getGlobalElement(local_indices[k]);

                        algebraic->matrices->B_2->replaceGlobalValues(lagrange_id,
                                                                      Teuchos::tuple<global_ordinal_type>(column_index),
                                                                      Teuchos::tuple<scalar>(values[k]));
                    }

                    algebraic->matrices->C->replaceGlobalValues(lagrange_id,
                                                                Teuchos::tuple<global_ordinal_type>(lagrange_id),
                                                                Teuchos::tuple<scalar>((scalar)0.));

                    algebraic->matrices->Cb->replaceGlobalValues(lagrange_id,
                                                                 Teuchos::tuple<global_ordinal_type>(lagrange_id),
                                                                 Teuchos::tuple<scalar>((scalar)1.));

                    if (numLMPN == 2 || numLMPN == 4)
                    {
                        algebraic->matrices->B->getLocalRowView(local_index * numLMPN + numLMPN - 1, local_indices, values);

                        for (auto k = 0; k < local_indices.size(); ++k)
                        {
                            global_ordinal_type column_index = algebraic->matrices->B->getColMap()->getGlobalElement(local_indices[k]);

                            algebraic->matrices->B_2->replaceGlobalValues(lagrange_id + numLMPN - 1,
                                                                          Teuchos::tuple<global_ordinal_type>(column_index),
                                                                          Teuchos::tuple<scalar>(values[k]));
                        }

                        algebraic->matrices->C->replaceGlobalValues(lagrange_id + numLMPN - 1,
                                                                    Teuchos::tuple<global_ordinal_type>(lagrange_id + numLMPN - 1),
                                                                    Teuchos::tuple<scalar>((scalar)0.));

                        algebraic->matrices->Cb->replaceGlobalValues(lagrange_id + numLMPN - 1,
                                                                     Teuchos::tuple<global_ordinal_type>(lagrange_id + numLMPN - 1),
                                                                     Teuchos::tuple<scalar>((scalar)1.));
                    }
                }
            }
            // If the current Lagrange multiplier is inactive for every sample of the current ensemble,
            // we can do the update without using the mask entries.
            else if (!MaskLogic::AND(current_mask))
            {
                Teuchos::ArrayView<const local_ordinal_type> local_indices;
                Teuchos::ArrayView<const scalar> values;

                if (domain->contactsList->isSticking(interface_id))
                {
                    for (size_t j = 0; j < numLMPN; ++j)
                    {
                        algebraic->matrices->B->getLocalRowView(local_index * numLMPN + j, local_indices, values);

                        for (auto k = 0; k < local_indices.size(); ++k)
                        {
                            global_ordinal_type column_index = algebraic->matrices->B->getColMap()->getGlobalElement(local_indices[k]);

                            algebraic->matrices->B_2->replaceGlobalValues(lagrange_id + j,
                                                                          Teuchos::tuple<global_ordinal_type>(column_index),
                                                                          Teuchos::tuple<scalar>((scalar)0.));
                        }

                        algebraic->matrices->C->replaceGlobalValues(lagrange_id + j,
                                                                    Teuchos::tuple<global_ordinal_type>(lagrange_id + j),
                                                                    Teuchos::tuple<scalar>((scalar)1.));

                        algebraic->matrices->Cb->replaceGlobalValues(lagrange_id + j,
                                                                     Teuchos::tuple<global_ordinal_type>(lagrange_id + j),
                                                                     Teuchos::tuple<scalar>((scalar)0.));
                    }
                }
                else
                {
                    algebraic->matrices->B->getLocalRowView(local_index * numLMPN, local_indices, values);

                    for (auto k = 0; k < local_indices.size(); ++k)
                    {
                        global_ordinal_type column_index = algebraic->matrices->B->getColMap()->getGlobalElement(local_indices[k]);

                        algebraic->matrices->B_2->replaceGlobalValues(lagrange_id,
                                                                      Teuchos::tuple<global_ordinal_type>(column_index),
                                                                      Teuchos::tuple<scalar>((scalar)0.));
                    }

                    algebraic->matrices->C->replaceGlobalValues(lagrange_id,
                                                                Teuchos::tuple<global_ordinal_type>(lagrange_id),
                                                                Teuchos::tuple<scalar>((scalar)1.));

                    algebraic->matrices->Cb->replaceGlobalValues(lagrange_id,
                                                                 Teuchos::tuple<global_ordinal_type>(lagrange_id),
                                                                 Teuchos::tuple<scalar>((scalar)0.));

                    if (numLMPN == 2 || numLMPN == 4)
                    {
                        algebraic->matrices->B->getLocalRowView(local_index * numLMPN + numLMPN - 1, local_indices, values);

                        for (auto k = 0; k < local_indices.size(); ++k)
                        {
                            global_ordinal_type column_index = algebraic->matrices->B->getColMap()->getGlobalElement(local_indices[k]);

                            algebraic->matrices->B_2->replaceGlobalValues(lagrange_id + numLMPN - 1,
                                                                          Teuchos::tuple<global_ordinal_type>(column_index),
                                                                          Teuchos::tuple<scalar>((scalar)0.));
                        }

                        algebraic->matrices->C->replaceGlobalValues(lagrange_id + numLMPN - 1,
                                                                    Teuchos::tuple<global_ordinal_type>(lagrange_id + numLMPN - 1),
                                                                    Teuchos::tuple<scalar>((scalar)1.));

                        algebraic->matrices->Cb->replaceGlobalValues(lagrange_id + numLMPN - 1,
                                                                     Teuchos::tuple<global_ordinal_type>(lagrange_id + numLMPN - 1),
                                                                     Teuchos::tuple<scalar>((scalar)0.));
                    }
                }
            }
            // Else, if the activity of the current Lagrange multiplier depends on the samples, we use the mask to do the update.
            else
            {

                Teuchos::ArrayView<const local_ordinal_type> local_indices;
                Teuchos::ArrayView<const scalar> values;

                if (domain->contactsList->isSticking(interface_id))
                {
                    for (size_t j = 0; j < numLMPN; ++j)
                    {
                        algebraic->matrices->B->getLocalRowView(local_index * numLMPN + j, local_indices, values);

                        for (auto k = 0; k < local_indices.size(); ++k)
                        {
                            global_ordinal_type column_index = algebraic->matrices->B->getColMap()->getGlobalElement(local_indices[k]);

                            algebraic->matrices->B_2->replaceGlobalValues(lagrange_id + j,
                                                                          Teuchos::tuple<global_ordinal_type>(column_index),
                                                                          Teuchos::tuple<scalar>(oneMask * values[k]));
                        }

                        algebraic->matrices->C->replaceGlobalValues(lagrange_id + j,
                                                                    Teuchos::tuple<global_ordinal_type>(lagrange_id + j),
                                                                    Teuchos::tuple<scalar>(oneNotMask));

                        algebraic->matrices->Cb->replaceGlobalValues(lagrange_id + j,
                                                                     Teuchos::tuple<global_ordinal_type>(lagrange_id + j),
                                                                     Teuchos::tuple<scalar>(oneMask));
                    }
                }
                else
                {
                    algebraic->matrices->B->getLocalRowView(local_index * numLMPN, local_indices, values);

                    for (auto k = 0; k < local_indices.size(); ++k)
                    {
                        global_ordinal_type column_index = algebraic->matrices->B->getColMap()->getGlobalElement(local_indices[k]);

                        algebraic->matrices->B_2->replaceGlobalValues(lagrange_id,
                                                                      Teuchos::tuple<global_ordinal_type>(column_index),
                                                                      Teuchos::tuple<scalar>(oneMask * values[k]));
                    }

                    algebraic->matrices->C->replaceGlobalValues(lagrange_id,
                                                                Teuchos::tuple<global_ordinal_type>(lagrange_id),
                                                                Teuchos::tuple<scalar>(oneNotMask));

                    algebraic->matrices->Cb->replaceGlobalValues(lagrange_id,
                                                                 Teuchos::tuple<global_ordinal_type>(lagrange_id),
                                                                 Teuchos::tuple<scalar>(oneMask));

                    if (numLMPN == 2 || numLMPN == 4)
                    {
                        algebraic->matrices->B->getLocalRowView(local_index * numLMPN + numLMPN - 1, local_indices, values);

                        for (auto k = 0; k < local_indices.size(); ++k)
                        {
                            global_ordinal_type column_index = algebraic->matrices->B->getColMap()->getGlobalElement(local_indices[k]);

                            algebraic->matrices->B_2->replaceGlobalValues(lagrange_id + numLMPN - 1,
                                                                          Teuchos::tuple<global_ordinal_type>(column_index),
                                                                          Teuchos::tuple<scalar>(oneMask * values[k]));
                        }

                        algebraic->matrices->C->replaceGlobalValues(lagrange_id + numLMPN - 1,
                                                                    Teuchos::tuple<global_ordinal_type>(lagrange_id + numLMPN - 1),
                                                                    Teuchos::tuple<scalar>(oneNotMask));

                        algebraic->matrices->Cb->replaceGlobalValues(lagrange_id + numLMPN - 1,
                                                                     Teuchos::tuple<global_ordinal_type>(lagrange_id + numLMPN - 1),
                                                                     Teuchos::tuple<scalar>(oneMask));
                    }
                }
            }
        }

        old_old_activity(local_index) = old_activity(local_index);
        old_activity(local_index) = oneMask;
    }

    algebraic->matrices->B_2->fillComplete(algebraic->map->mapDofs, algebraic->map->mapLagrangeDofs);
    algebraic->matrices->C->fillComplete();
    algebraic->matrices->Cb->fillComplete();

    RCP<tpetra_vector_type> initialActiveGap(new tpetra_vector_type(algebraic->map->mapLagrangeDofs, true));
    algebraic->matrices->Cb->apply(*algebraic->vectors->initialGap, *initialActiveGap);

    RCP<xpetra_mvector_type> xg =
        rcp(new xpetra_tmvector_type(initialActiveGap));
    algebraic->vectors->rhsBlockedMultiVector->setMultiVector(1, xg, true);
    algebraic->vectors->rhsMultiVector = algebraic->vectors->rhsBlockedMultiVector->Merge();

    size_t numtasks = algebraic->map->mapDofs->getComm()->getSize();

    double tmp;

    if (myRank == 0)
    {
        Mask<scalar> globalHasConverged = hasConverged;
        bool globalHasConverged_old = hasConverged_old;

        for (size_t i = 1; i < numtasks; ++i)
        {
            for (size_t l = 0; l < ensemble_size; ++l)
            {
                MPI_Recv(&tmp, 1, MPI_DOUBLE, i, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
                if (tmp < 0.)
                    globalHasConverged.set(l, false);
            }
            MPI_Recv(&tmp, 1, MPI_DOUBLE, i, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
            if (tmp < 0.)
                globalHasConverged_old = false;
        }
        hasConverged = globalHasConverged;
        hasConverged_old = globalHasConverged_old;

        for (size_t i = 1; i < numtasks; ++i)
        {
            for (size_t l = 0; l < ensemble_size; ++l)
            {
                tmp = hasConverged.get(l) ? 1. : -1.;
                MPI_Send(&tmp, 1, MPI_DOUBLE, i, 0, MPI_COMM_WORLD);
            }
            tmp = hasConverged_old ? 1. : -1.;
            MPI_Send(&tmp, 1, MPI_DOUBLE, i, 0, MPI_COMM_WORLD);
        }
    }
    else
    {
        for (size_t l = 0; l < ensemble_size; ++l)
        {
            tmp = hasConverged.get(l) ? 1. : -1.;
            MPI_Send(&tmp, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);
        }
        tmp = hasConverged_old ? 1. : -1.;
        MPI_Send(&tmp, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD);
        for (size_t l = 0; l < ensemble_size; ++l)
        {
            MPI_Recv(&tmp, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
            hasConverged.set(l, tmp > 0.);
        }
        MPI_Recv(&tmp, 1, MPI_DOUBLE, 0, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        hasConverged_old = (tmp > 0.);
    }

    if (hasConverged_old)
    {
        if (myRank == 0)
        {
            std::cout << "-----------------------" << std::endl;
            std::cout << "---Has not converged---" << std::endl;
            std::cout << "-----------------------" << std::endl;
        }
        return true;
    }

    return hasConverged;
}

/**
 * @brief Create the block matrix.
 * 
 * Arguments:
 * <ul>
 *  <li> gamma: the value used for the Augmented Lagrangian formulation,
 *  <li> write: a boolean which is used to write matrices on disk or not.
 * </ul>
 */
template <typename scalar>
void DiscreteProblem<scalar>::createBlockMatrix(scalar gamma, bool write, bool merge)
{
    using Teuchos::Array;
    using Teuchos::RCP;
    using Teuchos::rcp;

    const std::vector<RCP<const tpetra_map_type>> maps = {algebraic->map->mapDofs, algebraic->map->mapLagrangeDofs};

    const size_t numDofs = algebraic->map->mapDofs->getGlobalNumElements();
    const size_t numMyDofs = algebraic->map->mapDofs->getNodeNumElements();
    const size_t numMyDofsWO = algebraic->map->mapDofsWO->getNodeNumElements();
    Teuchos::ArrayView<const global_ordinal_type> myDofs = algebraic->map->mapDofs->getNodeElementList();
    Teuchos::ArrayView<const global_ordinal_type> myDofsWO = algebraic->map->mapDofsWO->getNodeElementList();

    const size_t numLagrangeDofs = algebraic->map->mapLagrangeDofs->getGlobalNumElements();
    const size_t numMyLagrangeDofs = algebraic->map->mapLagrangeDofs->getNodeNumElements();
    Teuchos::ArrayView<const global_ordinal_type> myLagrangeDofs = algebraic->map->mapLagrangeDofs->getNodeElementList();

    Array<global_ordinal_type> myFullDofs(numMyDofs + numMyLagrangeDofs);
    for (auto i = 0; i < numMyDofs; ++i)
        myFullDofs[i] = myDofs[i];
    for (auto i = 0; i < numMyLagrangeDofs; ++i)
        myFullDofs[numMyDofs + i] = numDofs + myLagrangeDofs[i];

    Array<global_ordinal_type> myFullDofsWO(numMyDofsWO + numMyLagrangeDofs);
    for (auto i = 0; i < numMyDofsWO; ++i)
        myFullDofsWO[i] = myDofsWO[i];
    for (auto i = 0; i < numMyLagrangeDofs; ++i)
        myFullDofsWO[numMyDofsWO + i] = numDofs + myLagrangeDofs[i];

    algebraic->map->fullmap =
        rcp(new tpetra_map_type(numDofs + numLagrangeDofs,
                                myFullDofs,
                                algebraic->map->mapDofs->getIndexBase(),
                                algebraic->map->mapDofs->getComm()));
    algebraic->map->fullmapWO =
        rcp(new tpetra_map_type(numDofs + numLagrangeDofs,
                                myFullDofsWO,
                                algebraic->map->mapDofs->getIndexBase(),
                                algebraic->map->mapDofs->getComm()));

    RCP<xpetra_map_type> xfullmap = rcp(new xpetra_tmap_type(algebraic->map->fullmap));

    const size_t myRank = algebraic->map->mapDofs->getComm()->getRank();
    const size_t numLclIndices = (myRank == 0) ? numDofs + numLagrangeDofs : 0;

    algebraic->map->fullmapOutput =
        rcp(new tpetra_map_type(numDofs + numLagrangeDofs,
                                numLclIndices,
                                algebraic->map->mapDofs->getIndexBase(),
                                algebraic->map->mapDofs->getComm()));

    std::vector<RCP<const xpetra_map_type>> xsubmaps = {rcp(new xpetra_tmap_type(algebraic->map->mapDofs)), //algebraic->map->thermomecBlockedMap,
                                                        rcp(new xpetra_tmap_type(algebraic->map->mapLagrangeDofs))};

    algebraic->map->blockedMap =
        rcp(new xpetra_bmap_type(xfullmap, xsubmaps, true));

    algebraic->matrices->blockedMatrix =
        rcp(new xpetra_bcrs_type(algebraic->map->blockedMap, algebraic->map->blockedMap, 8));

    //Update B based on the initial activity

    algebraic->matrices->B_2->resumeFill();
    algebraic->matrices->C->resumeFill();
    algebraic->matrices->Cb->resumeFill();

    const size_t numDPN = algebraic->map->numPrimalDPN;

    size_t numLMPN;
    if (domain->contactsList->hasAtLeastOneSticking() && numDPN == 4)
        numLMPN = 4;
    else if (domain->contactsList->hasAtLeastOneSticking())
        numLMPN = 3;
    else if (numDPN == 4)
        numLMPN = 2;
    else
        numLMPN = 1;

    const int numMyLagrangeNum = algebraic->map->mapLagrangeDofs->getNodeNumElements();

    const int numMyMechLagrangeNum = numMyLagrangeNum / numLMPN;

    old_activity = Kokkos::View<scalar *, Kokkos::LayoutRight>("R", numMyMechLagrangeNum);
    old_old_activity = Kokkos::View<scalar *, Kokkos::LayoutRight>("R", numMyMechLagrangeNum);

    for (auto local_index = 0; local_index < numMyMechLagrangeNum; ++local_index)
    {
        global_ordinal_type lagrange_id = algebraic->map->mapLagrangeDofs->getGlobalElement(local_index);
        global_ordinal_type node_id = algebraic->map->lm_to_dof[lagrange_id];
        lagrange_id = lagrange_id * numLMPN;
        local_ordinal_type interface_id = domain->contactsList->getInterfaceOfSlaveNode(node_id);
        if (domain->contactsList->isInitiallyOpen(interface_id))
        {
            if (domain->contactsList->isNodeInitiallyClosed(interface_id, node_id))
            {
                old_activity(local_index) = (scalar)1.;
                continue;
            }
            old_activity(local_index) = (scalar)0.;

            Teuchos::ArrayView<const local_ordinal_type> local_indices;
            Teuchos::ArrayView<const scalar> values;

            for (size_t j = 0; j < numLMPN; ++j)
            {
                algebraic->matrices->B_2->getLocalRowView(local_index * numLMPN + j, local_indices, values);

                for (auto k = 0; k < local_indices.size(); ++k)
                {
                    global_ordinal_type column_index = algebraic->matrices->B->getColMap()->getGlobalElement(local_indices[k]);

                    algebraic->matrices->B_2->replaceGlobalValues(lagrange_id + j,
                                                                  Teuchos::tuple<global_ordinal_type>(column_index),
                                                                  Teuchos::tuple<scalar>((scalar)0.));
                }

                algebraic->matrices->C->replaceGlobalValues(lagrange_id + j,
                                                            Teuchos::tuple<global_ordinal_type>(lagrange_id + j),
                                                            Teuchos::tuple<scalar>((scalar)1.));

                algebraic->matrices->Cb->replaceGlobalValues(lagrange_id + j,
                                                             Teuchos::tuple<global_ordinal_type>(lagrange_id + j),
                                                             Teuchos::tuple<scalar>((scalar)0.));
            }
        }
        else
        {
            old_activity(local_index) = (scalar)1.;
            if (domain->contactsList->isSticking(interface_id) == false && domain->contactsList->hasAtLeastOneSticking() == true)
            {
                Teuchos::ArrayView<const local_ordinal_type> local_indices;
                Teuchos::ArrayView<const scalar> values;

                for (size_t j = 1; j < numLMPN; ++j)
                {
                    algebraic->matrices->B_2->getLocalRowView(local_index * numLMPN + j, local_indices, values);

                    for (auto k = 0; k < local_indices.size(); ++k)
                    {
                        global_ordinal_type column_index = algebraic->matrices->B->getColMap()->getGlobalElement(local_indices[k]);

                        algebraic->matrices->B_2->replaceGlobalValues(lagrange_id + j,
                                                                      Teuchos::tuple<global_ordinal_type>(column_index),
                                                                      Teuchos::tuple<scalar>((scalar)0.));
                    }

                    algebraic->matrices->C->replaceGlobalValues(lagrange_id + j,
                                                                Teuchos::tuple<global_ordinal_type>(lagrange_id + j),
                                                                Teuchos::tuple<scalar>((scalar)1.));

                    algebraic->matrices->Cb->replaceGlobalValues(lagrange_id + j,
                                                                 Teuchos::tuple<global_ordinal_type>(lagrange_id + j),
                                                                 Teuchos::tuple<scalar>((scalar)0.));
                }
            }
        }
    }
    algebraic->matrices->B_2->fillComplete(algebraic->map->mapDofs, algebraic->map->mapLagrangeDofs);
    algebraic->matrices->C->fillComplete();
    algebraic->matrices->Cb->fillComplete();

    algebraic->matrices->xB = rcp(new xpetra_tcrs_type(algebraic->matrices->B_2));
    algebraic->matrices->xB_T = rcp(new xpetra_tcrs_type(algebraic->matrices->B_T));
    algebraic->matrices->xC = rcp(new xpetra_tcrs_type(algebraic->matrices->C));

    algebraic->matrices->xwB = rcp(new xpetra_wcrs_type(algebraic->matrices->xB));
    algebraic->matrices->xwB_T = rcp(new xpetra_wcrs_type(algebraic->matrices->xB_T));
    algebraic->matrices->xwC = rcp(new xpetra_wcrs_type(algebraic->matrices->xC));

    RCP<Xpetra::Matrix<scalar, local_ordinal_type, global_ordinal_type, node_type>> CTC = Teuchos::null;
    RCP<Xpetra::Matrix<scalar, local_ordinal_type, global_ordinal_type, node_type>> ApCTC = Teuchos::null;
    std::string MatrixMatrix_output_file = "MatrixMatrix_out.txt";
    std::ofstream MatrixMatrix_ofstream(MatrixMatrix_output_file);
    RCP<Teuchos::FancyOStream> verbOut = Teuchos::getFancyOStream(Teuchos::rcpFromRef(MatrixMatrix_ofstream));
    CTC =
        Xpetra::MatrixMatrix<scalar, local_ordinal_type, global_ordinal_type, node_type>::Multiply(*(algebraic->matrices->xwB_T),
                                                                                                   false,
                                                                                                   *(algebraic->matrices->xwB),
                                                                                                   false,
                                                                                                   *verbOut);
    RCP<xpetra_bcrs_type> bA = Teuchos::rcp_dynamic_cast<xpetra_bcrs_type>(algebraic->matrices->A);

    if (gamma > 0.)
    {
        if (bA != Teuchos::null && merge)
            Xpetra::MatrixMatrix<scalar, local_ordinal_type, global_ordinal_type, node_type>::TwoMatrixAdd(*CTC,
                                                                                                           false,
                                                                                                           gamma,
                                                                                                           *(bA->Merge()),
                                                                                                           false,
                                                                                                           (scalar)1.,
                                                                                                           ApCTC,
                                                                                                           *verbOut);
        else
            Xpetra::MatrixMatrix<scalar, local_ordinal_type, global_ordinal_type, node_type>::TwoMatrixAdd(*CTC,
                                                                                                           false,
                                                                                                           gamma,
                                                                                                           *(algebraic->matrices->A),
                                                                                                           false,
                                                                                                           (scalar)1.,
                                                                                                           ApCTC,
                                                                                                           *verbOut);
    }
    else
    {
        if (bA != Teuchos::null && merge)
            ApCTC = bA->Merge();
        else
            ApCTC = algebraic->matrices->A;
    }

    algebraic->matrices->blockedMatrix->setMatrix(0, 0, ApCTC);
    algebraic->matrices->blockedMatrix->setMatrix(0, 1, algebraic->matrices->xwB_T);
    algebraic->matrices->blockedMatrix->setMatrix(1, 0, algebraic->matrices->xwB);
    algebraic->matrices->blockedMatrix->setMatrix(1, 1, algebraic->matrices->xwC);

    algebraic->matrices->blockedMatrix->fillComplete();
    RCP<const xpetra_wcrs_type> tmpOp = Teuchos::rcp_dynamic_cast<const xpetra_wcrs_type>(ApCTC);
    const RCP<const xpetra_tcrs_type> &tmp_ECrsMtx = Teuchos::rcp_dynamic_cast<const xpetra_tcrs_type>(tmpOp->getCrsMatrix());
    RCP<const tpetra_crs_type> tpetra_ApCTC = tmp_ECrsMtx->getTpetra_CrsMatrix();

    if (write)
        Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeSparseFile(std::string("ApCTC_mm.txt"), tpetra_ApCTC);
}

/**
 * @brief Create the block vectors.
 * 
 * Argument:
 * <ul>
 *  <li> write: a boolean which is used to write vectors on disk or not.
 * </ul>
 */
template <typename scalar>
void DiscreteProblem<scalar>::createBlockMVector(bool write)
{
    using Teuchos::Array;
    using Teuchos::RCP;
    using Teuchos::rcp;

    RCP<tpetra_vector_type> intialContactPosition(new tpetra_vector_type(algebraic->map->mapDofs, true));
    algebraic->vectors->initialGap = rcp(new tpetra_vector_type(algebraic->map->mapLagrangeDofs, true));
    RCP<tpetra_vector_type> initialActiveGap(new tpetra_vector_type(algebraic->map->mapLagrangeDofs, true));

    algebraic->vectors->lagrange = rcp(new tpetra_vector_type(algebraic->map->mapLagrangeDofs, true));

    {
        size_t numMyContacts = domain->contactsList->getContactNumber();
        size_t numPrimalDPN = algebraic->map->numPrimalDPN;
        size_t numMPrimalDPN = 3;
        size_t offSet = 0;
        if (numPrimalDPN == 4)
            offSet = algebraic->map->mapNodes->getGlobalNumElements();

        for (auto i = 0; i < numMyContacts; ++i)
        {
            size_t numMySlaveNodes = domain->contactsList->getSlaveNodesSize(i);
            for (auto j = 0; j < numMySlaveNodes; ++j)
            {
                // get unique node id
                auto global_nodeid = domain->contactsList->getSlaveNode(i, j);
                if (algebraic->map->mapNodes->isNodeGlobalElement(global_nodeid))
                {
                    auto local_nodeid = algebraic->map->mapNodesWO->getLocalElement(global_nodeid);
                    for (auto k = 0; k < 3; ++k)
                    {
                        scalar tmp = domain->nodesList->nodes(local_nodeid, k) +
                                     domain->contactsList->getSlaveNormal(i, k) *
                                         loads->preloadList->getPreloadValue(i);
                        intialContactPosition->replaceGlobalValue(offSet + global_nodeid * numMPrimalDPN + k, tmp);
                    }
                }
            }

            size_t numMyMasterNodes = domain->contactsList->getMasterNodesSize(i);
            for (auto j = 0; j < numMyMasterNodes; ++j)
            {
                // get unique node id
                auto global_nodeid = domain->contactsList->getMasterNode(i, j);
                if (algebraic->map->mapNodes->isNodeGlobalElement(global_nodeid))
                {
                    auto local_nodeid = algebraic->map->mapNodesWO->getLocalElement(global_nodeid);
                    for (auto k = 0; k < 3; ++k)
                        intialContactPosition->replaceGlobalValue(offSet + global_nodeid * numMPrimalDPN + k,
                                                                  ((scalar)domain->nodesList->nodes(local_nodeid, k)));
                }
            }
        }
    }

    algebraic->matrices->B_G->apply(*intialContactPosition, *algebraic->vectors->initialGap, Teuchos::NO_TRANS, ((scalar)-1.));

    algebraic->matrices->Cb->apply(*algebraic->vectors->initialGap, *initialActiveGap);

    // TODO: fix me when there is mesh tying and contact together!
    if (domain->contactsList->isTying(0))
        initialActiveGap->scale((scalar)0.);

    if (write)
        Tpetra::MatrixMarket::Writer<tpetra_crs_type>::writeDenseFile(std::string("initialActiveGap_mm.txt"), initialActiveGap);

    RCP<xpetra_mvector_type> xb =
        rcp(new xpetra_tmvector_type(algebraic->vectors->b));

    RCP<xpetra_mvector_type> xg =
        rcp(new xpetra_tmvector_type(initialActiveGap));

    RCP<xpetra_mvector_type> xx =
        rcp(new xpetra_tmvector_type(algebraic->vectors->x));

    RCP<xpetra_mvector_type> xlagrange =
        rcp(new xpetra_tmvector_type(algebraic->vectors->lagrange));

    algebraic->vectors->rhsBlockedMultiVector =
        rcp(new xpetra_bmvector_type(algebraic->map->blockedMap, 1, true));

    algebraic->vectors->solutionBlockedMultiVector =
        rcp(new xpetra_bmvector_type(algebraic->map->blockedMap, 1, true));

    algebraic->vectors->rhsBlockedMultiVector->setMultiVector(0, xb, true);
    algebraic->vectors->rhsBlockedMultiVector->setMultiVector(1, xg, true);

    algebraic->vectors->solutionBlockedMultiVector->setMultiVector(0, xx, true);
    algebraic->vectors->solutionBlockedMultiVector->setMultiVector(1, xlagrange, true);

    algebraic->vectors->rhsMultiVector = algebraic->vectors->rhsBlockedMultiVector->Merge();
    algebraic->vectors->solutionMultiVector = algebraic->vectors->solutionBlockedMultiVector->Merge();
}

/**
 * @brief Compute the contacts matrices.
 * 
 * Argument:
 * <ul>
 *  <li> comm: the used MPI communicator.
 * </ul>
 */
template <typename scalar>
void DiscreteProblem<scalar>::computeContactMatrices(MPI_Comm comm)
{
    compute_B<scalar>(algebraic->matrices,
                      domain->elementsList,
                      domain->nodesList,
                      domain->contactsList,
                      algebraic->map, comm);
}

/**
 * @brief Compute the primal matrix.
 */
template <typename scalar>
void DiscreteProblem<scalar>::computeMatrices(void)
{
    using Teuchos::RCP;
    using Teuchos::rcp;

    typedef KokkosSparse::CrsMatrix<scalar, local_ordinal_type, typename tpetra_crs_type::execution_space> local_matrix_type;

    typedef Kokkos::View<scalar ***, Kokkos::LayoutRight, Kokkos::DefaultExecutionSpace> elem_matrix_view_type;

    local_matrix_type local_L, local_K, local_S;

    if (numPrimalDPN == 4 || numPrimalDPN == 1)
    {
        algebraic->matrices->L->resumeFill();
        algebraic->matrices->L->setAllToScalar((scalar)0.0);

        local_L = algebraic->matrices->L->getLocalMatrix();
    }
    if (numPrimalDPN == 4 || numPrimalDPN == 3)
    {
        algebraic->matrices->K->resumeFill();
        algebraic->matrices->K->setAllToScalar((scalar)0.0);

        local_K = algebraic->matrices->K->getLocalMatrix();
    }
    if (numPrimalDPN == 4)
    {
        algebraic->matrices->S->resumeFill();
        algebraic->matrices->S->setAllToScalar((scalar)0.0);

        local_S = algebraic->matrices->S->getLocalMatrix();
    }

    auto numMyElems = domain->elementsList->getElementNumber();

    elem_matrix_view_type K, L, S;

#ifdef KOKKOS_ENABLE_DEPRECATED_CODE
    int pool_size = Kokkos::DefaultExecutionSpace::thread_pool_size();
#else
    int pool_size = Kokkos::DefaultExecutionSpace::impl_thread_pool_size();
#endif
    int max_element_size = 1;

    for (int i_elem = 0; i_elem < numMyElems; ++i_elem)
    {
        if (domain->elementsList->getElementType(i_elem) == static_cast<int>(tbox::ELTYPE::HEX8))
        {
            max_element_size = 8;
            break;
        }
        if (domain->elementsList->getElementType(i_elem) == static_cast<int>(tbox::ELTYPE::TETRA4))
        {
            max_element_size = 4;
            break;
        }
    }

    K = Kokkos::View<scalar ***, Kokkos::LayoutRight, Kokkos::DefaultExecutionSpace>("A",
                                                                                     pool_size,
                                                                                     3 * max_element_size,
                                                                                     3 * max_element_size);
    L = Kokkos::View<scalar ***, Kokkos::LayoutRight, Kokkos::DefaultExecutionSpace>("A",
                                                                                     pool_size,
                                                                                     max_element_size,
                                                                                     max_element_size);
    S = Kokkos::View<scalar ***, Kokkos::LayoutRight, Kokkos::DefaultExecutionSpace>("A",
                                                                                     pool_size,
                                                                                     3 * max_element_size,
                                                                                     max_element_size);

    ElementMatrices<scalar, static_cast<int>(tbox::ELTYPE::HEX8), 8> *elementMatrices_HEX8 =
        new ElementMatrices<scalar, static_cast<int>(tbox::ELTYPE::HEX8), 8>(domain, numPrimalDPN, pool_size);
    ElementMatrices<scalar, static_cast<int>(tbox::ELTYPE::TETRA4), 4> *elementMatrices_TETRA4 =
        new ElementMatrices<scalar, static_cast<int>(tbox::ELTYPE::TETRA4), 4>(domain, numPrimalDPN, pool_size);

    Kokkos::parallel_for(
        numMyElems, KOKKOS_LAMBDA(const int i_elem) {
            bool bContinue = true;

#ifdef KOKKOS_ENABLE_DEPRECATED_CODE
            int i_thread = Kokkos::DefaultExecutionSpace::thread_pool_rank();
#else
            int i_thread = Kokkos::DefaultExecutionSpace::impl_thread_pool_rank();
#endif
            const int e_size = domain->elementsList->getElementSize(i_elem);

            auto K_e = subview(K, i_thread, make_pair(0, 3 * e_size), make_pair(0, 3 * e_size));
            auto L_e = subview(L, i_thread, make_pair(0, e_size), make_pair(0, e_size));
            auto S_e = subview(S, i_thread, make_pair(0, 3 * e_size), make_pair(0, e_size));

            if (domain->elementsList->getElementType(i_elem) == static_cast<int>(tbox::ELTYPE::HEX8))
            {
                elementMatrices_HEX8->compute(i_elem, i_thread);
                Kokkos::deep_copy(K_e, elementMatrices_HEX8->getK(i_thread));
                Kokkos::deep_copy(L_e, elementMatrices_HEX8->getL(i_thread));
                Kokkos::deep_copy(S_e, elementMatrices_HEX8->getS(i_thread));
            }

            else if (domain->elementsList->getElementType(i_elem) == static_cast<int>(tbox::ELTYPE::TETRA4))
            {
                elementMatrices_TETRA4->compute(i_elem, i_thread);
                Kokkos::deep_copy(K_e, elementMatrices_TETRA4->getK(i_thread));
                Kokkos::deep_copy(L_e, elementMatrices_TETRA4->getL(i_thread));
                Kokkos::deep_copy(S_e, elementMatrices_TETRA4->getS(i_thread));
            }
            else
                bContinue = false;

            if (bContinue)
                for (auto i = 0; i < e_size; ++i)
                {
                    local_ordinal_type node_i = domain->elementsList->getElementNode(i_elem, i);
                    global_ordinal_type global_node_id_i = algebraic->map->mapNodesWO->getGlobalElement(node_i);
                    // Test if mapNodes ownes global_node_id_i
                    if (algebraic->map->mapNodes->isNodeGlobalElement(global_node_id_i))
                    {
                        local_ordinal_type local_node_id_i = algebraic->map->mapNodes->getLocalElement(global_node_id_i);
                        for (auto j = 0; j < e_size; ++j)
                        {
                            local_ordinal_type node_j = domain->elementsList->getElementNode(i_elem, j);

                            if (numPrimalDPN == 4 || numPrimalDPN == 1)
                            {
                                local_ordinal_type cols = node_j;
                                scalar vals = L_e(i, j);

                                local_L.sumIntoValues(local_node_id_i, &cols, 1, &vals, false, true);
                            }
                            if (numPrimalDPN == 4 || numPrimalDPN == 3)
                            {
                                for (auto k = 0; k < 3; ++k)
                                    for (auto l = 0; l < 3; ++l)
                                    {
                                        local_ordinal_type cols = 3 * node_j + l;
                                        scalar vals = K_e(3 * i + k, 3 * j + l);
                                        local_K.sumIntoValues(3 * local_node_id_i + k, &cols, 1, &vals, false, true);
                                    }
                            }
                            if (numPrimalDPN == 4)
                            {
                                for (auto k = 0; k < 3; ++k)
                                {
                                    local_ordinal_type cols = node_j;
                                    scalar vals = S_e(3 * i + k, j);
                                    local_S.sumIntoValues(3 * local_node_id_i + k, &cols, 1, &vals, false, true);
                                }
                            }
                        }
                    }
                }
        });

    if (numPrimalDPN == 4 || numPrimalDPN == 1)
    {
        algebraic->matrices->L->fillComplete();
    }
    if (numPrimalDPN == 4 || numPrimalDPN == 3)
    {
        algebraic->matrices->K->fillComplete();
    }
    if (numPrimalDPN == 4)
    {
        algebraic->matrices->S->fillComplete();
    }

    delete elementMatrices_HEX8;
    delete elementMatrices_TETRA4;

    if (numPrimalDPN == 4)
    {
        RCP<xpetra_bcrs_type> bA = rcp(new xpetra_bcrs_type(algebraic->map->thermomecBlockedMap, algebraic->map->thermomecBlockedMap, 8));

        RCP<xpetra_crs_type> xL = rcp(new xpetra_tcrs_type(algebraic->matrices->L));
        RCP<xpetra_crs_type> xK = rcp(new xpetra_tcrs_type(algebraic->matrices->K));
        RCP<xpetra_crs_type> xS = rcp(new xpetra_tcrs_type(algebraic->matrices->S));

        RCP<xpetra_wcrs_type> xwL = rcp(new xpetra_wcrs_type(xL));
        RCP<xpetra_wcrs_type> xwK = rcp(new xpetra_wcrs_type(xK));
        RCP<xpetra_wcrs_type> xwS = rcp(new xpetra_wcrs_type(xS));

        bA->setMatrix(0, 0, xwL);
        bA->setMatrix(1, 0, xwS);
        bA->setMatrix(1, 1, xwK);

        bA->fillComplete();

        algebraic->matrices->A = bA;
    }
    else if (numPrimalDPN == 3)
    {
        RCP<xpetra_crs_type> xK = rcp(new xpetra_tcrs_type(algebraic->matrices->K));
        algebraic->matrices->A = rcp(new xpetra_wcrs_type(xK));
    }
    else
    {
        RCP<xpetra_crs_type> xL = rcp(new xpetra_tcrs_type(algebraic->matrices->L));
        algebraic->matrices->A = rcp(new xpetra_wcrs_type(xL));
    }
}

/**
 * @brief Compute the rhs.
 * 
 * Argument:
 * <ul>
 *  <li> scaled: a boolean which describes whether the system is scaled or not.
 * </ul>
 */
template <typename scalar>
void DiscreteProblem<scalar>::computeLoads(bool scaled)
{

    using Teuchos::RCP;
    using Teuchos::rcp;
    using Teuchos::tuple;

    typedef Kokkos::View<scalar ***, Kokkos::LayoutRight, Kokkos::DefaultExecutionSpace> elem_matrix_view_type;

    elem_matrix_view_type N, N_S;

    auto numMyElems = domain->elementsList->getElementNumber();

#ifdef KOKKOS_ENABLE_DEPRECATED_CODE
    int pool_size = Kokkos::DefaultExecutionSpace::thread_pool_size();
#else
    int pool_size = Kokkos::DefaultExecutionSpace::impl_thread_pool_size();
#endif
    int max_element_size = 1;
    int max_element_size_S = 1;

    for (int i_elem = 0; i_elem < numMyElems; ++i_elem)
    {
        if (domain->elementsList->getElementType(i_elem) == static_cast<int>(tbox::ELTYPE::HEX8))
        {
            max_element_size = 8;
            max_element_size_S = 4;
            break;
        }
        if (domain->elementsList->getElementType(i_elem) == static_cast<int>(tbox::ELTYPE::TETRA4))
        {
            max_element_size = 4;
            max_element_size_S = 3;
            break;
        }
    }
    N = elem_matrix_view_type("A", pool_size, max_element_size, max_element_size);
    N_S = elem_matrix_view_type("A", pool_size, max_element_size_S, max_element_size_S);

    ElementVectors<scalar, static_cast<int>(tbox::ELTYPE::TRI3), 3> *elementVectors_TRI3 =
        new ElementVectors<scalar, static_cast<int>(tbox::ELTYPE::TRI3), 3>(domain, numPrimalDPN, pool_size);
    ElementVectors<scalar, static_cast<int>(tbox::ELTYPE::QUAD4), 4> *elementVectors_QUAD4 =
        new ElementVectors<scalar, static_cast<int>(tbox::ELTYPE::QUAD4), 4>(domain, numPrimalDPN, pool_size);
    ElementVectors<scalar, static_cast<int>(tbox::ELTYPE::TETRA4), 4> *elementVectors_TETRA4 =
        new ElementVectors<scalar, static_cast<int>(tbox::ELTYPE::TETRA4), 4>(domain, numPrimalDPN, pool_size);
    ElementVectors<scalar, static_cast<int>(tbox::ELTYPE::HEX8), 8> *elementVectors_HEX8 =
        new ElementVectors<scalar, static_cast<int>(tbox::ELTYPE::HEX8), 8>(domain, numPrimalDPN, pool_size);

    RCP<tpetra_vector_type> bWO = rcp(new tpetra_vector_type(algebraic->map->mapDofsWO, true));

    auto bWO_view = bWO->getLocalViewHost();

    // ------------------------------------------
    //
    //          volumetric heat sources
    //
    // ------------------------------------------

    auto numMySources = loads->sourcesList->getSourceNumber();

    for (auto i = 0; i < numMySources; ++i)
    {
        auto currentSourceSize = loads->sourcesList->getSourceSize(i);
        Kokkos::parallel_for(
            currentSourceSize, KOKKOS_LAMBDA(const int e1) {
                local_ordinal_type i_elem = loads->sourcesList->getSourceElement(i, e1);
#ifdef KOKKOS_ENABLE_DEPRECATED_CODE
                int i_thread = Kokkos::DefaultExecutionSpace::thread_pool_rank();
#else
                int i_thread = Kokkos::DefaultExecutionSpace::impl_thread_pool_rank();
#endif
                const int e_size = domain->elementsList->getElementSize(i_elem);

                auto N_e = subview(N, i_thread, make_pair(0, e_size), make_pair(0, e_size));

                bool bContinue = true;

                if (domain->elementsList->getElementType(i_elem) == static_cast<int>(tbox::ELTYPE::HEX8))
                {
                    elementVectors_HEX8->compute(i_elem, i_thread);
                    Kokkos::deep_copy(N_e, elementVectors_HEX8->getN(i_thread));
                }
                else if (domain->elementsList->getElementType(i_elem) == static_cast<int>(tbox::ELTYPE::TETRA4))
                {
                    elementVectors_TETRA4->compute(i_elem, i_thread);
                    Kokkos::deep_copy(N_e, elementVectors_TETRA4->getN(i_thread));
                }
                else
                    bContinue = false;

                if (bContinue)
                {
                    for (auto ii = 0; ii < e_size; ++ii)
                    {
                        local_ordinal_type local_i = domain->elementsList->getElementNode(i_elem, ii);
                        global_ordinal_type global_i = algebraic->map->mapNodesWO->getGlobalElement(local_i);

                        // Test if mapDofs ownes global_i
                        if (algebraic->map->mapNodes->isNodeGlobalElement(global_i))
                        {
                            scalar tmp = 0.;

                            for (auto jj = 0; jj < e_size; ++jj)
                                tmp += N_e(ii, jj);

                            tmp *= loads->sourcesList->getSourceValue(i);

                            Kokkos::atomic_fetch_add(&bWO_view(local_i, 0), tmp);
                        }
                    }
                }
            });
    }

    // ------------------------------------------
    //
    //       von Neumann boundary conditions
    //
    // ------------------------------------------

    auto numMynBCs = loads->neumannList->getNeumannNumber();

    const size_t numNodes = algebraic->map->mapNodes->getGlobalNumElements();
    //const size_t numMyNodes = algebraic->map->mapNodes->getNodeNumElements();
    const size_t numMyNodesWO = algebraic->map->mapNodesWO->getNodeNumElements();

    for (auto i = 0; i < numMynBCs; ++i)
    {
        auto currentBCSize = loads->neumannList->getNeumannSize(i);
        Kokkos::parallel_for(
            currentBCSize, KOKKOS_LAMBDA(const int e1) {
                local_ordinal_type i_elem = loads->neumannList->getNeumannElement(i, e1);

#ifdef KOKKOS_ENABLE_DEPRECATED_CODE
                int i_thread = Kokkos::DefaultExecutionSpace::thread_pool_rank();
#else
                int i_thread = Kokkos::DefaultExecutionSpace::impl_thread_pool_rank();
#endif
                const int e_size = domain->elementsList->getElementSize(i_elem);

                auto N_e = subview(N_S, i_thread, make_pair(0, e_size), make_pair(0, e_size));

                bool bContinue = true;

                if (domain->elementsList->getElementType(i_elem) == static_cast<int>(tbox::ELTYPE::TRI3))
                {
                    elementVectors_TRI3->compute(i_elem, i_thread);
                    Kokkos::deep_copy(N_e, elementVectors_TRI3->getN(i_thread));
                }
                else if (domain->elementsList->getElementType(i_elem) == static_cast<int>(tbox::ELTYPE::QUAD4))
                {
                    elementVectors_QUAD4->compute(i_elem, i_thread);
                    Kokkos::deep_copy(N_e, elementVectors_QUAD4->getN(i_thread));
                }
                else
                    bContinue = false;

                if (bContinue)
                {
                    for (auto j = 0; j < numPrimalDPN; ++j)
                    {
                        scalar which_dof = loads->neumannList->getNeumannDof(i, j);
                        scalar value = loads->neumannList->getNeumannValue(i, j);
                        if (which_dof == (scalar)1)
                        {
                            for (auto ii = 0; ii < e_size; ++ii)
                            {
                                local_ordinal_type local_i;
                                if (j == 3)
                                    local_i = domain->elementsList->getElementNode(i_elem, ii);
                                else
                                {
                                    if (numPrimalDPN == 4)
                                        local_i = numMyNodesWO + 3 * domain->elementsList->getElementNode(i_elem, ii) + j;
                                    else
                                        local_i = 3 * domain->elementsList->getElementNode(i_elem, ii) + j;
                                }

                                scalar tmp = 0.;

                                for (auto jj = 0; jj < e_size; ++jj)
                                    tmp += N_e(ii, jj);

                                tmp *= value;

                                Kokkos::atomic_fetch_add(&bWO_view(local_i, 0), tmp);
                            }
                        }
                    }
                }
            });
    }

    // ------------------------------------------
    //
    //        Dirichlet boundary conditions
    //
    // ------------------------------------------

    Tpetra::Export<> exportBWO(algebraic->map->mapDofsWO, algebraic->map->mapDofs);
    algebraic->vectors->b->doExport(*bWO, exportBWO, Tpetra::ADD);

    if (numPrimalDPN == 4 || numPrimalDPN == 3)
        algebraic->matrices->K->resumeFill();
    if (numPrimalDPN == 4 || numPrimalDPN == 1)
        algebraic->matrices->L->resumeFill();
    if (numPrimalDPN == 4)
        algebraic->matrices->S->resumeFill();

    auto numMyBCs = loads->dirichletList->getDirichletNumber();
    for (auto i = 0; i < numMyBCs; ++i)
    {
        auto currentBCSize = loads->dirichletList->getDirichletSize(i);
        for (int i_node = 0; i_node < currentBCSize; ++i_node)
        {
            global_ordinal_type n1 = loads->dirichletList->getDirichletNode(i, i_node);

            if (algebraic->map->mapNodes->isNodeGlobalElement(n1))
                for (auto j = 0; j < numPrimalDPN; ++j)
                {
                    scalar which_dof = loads->dirichletList->getDirichletDof(i, j);

                    if (which_dof == (scalar)1)
                    {
                        global_ordinal_type global_index, global_index_0, global_index_r0;

                        bool thermo = false;

                        if (j == 3)
                        {
                            global_index_0 = 0;
                            global_index_r0 = n1;

                            thermo = true;
                        }
                        else
                        {
                            if (numPrimalDPN == 4)
                                global_index_0 = numNodes;
                            else
                                global_index_0 = 0;
                            global_index_r0 = 3 * n1 + j;
                        }

                        global_index = global_index_0 + global_index_r0;

                        Teuchos::Array<global_ordinal_type> global_indices;

                        if (thermo)
                        {
                            size_t numColInds = algebraic->graph->L->getNumEntriesInGlobalRow(global_index);
                            global_indices.resize(numColInds);
                            algebraic->graph->L->getGlobalRowCopy(global_index, global_indices(), numColInds);

                            // Replace the initialized elements by zero
                            for (auto k = 0; k < numColInds; ++k)
                                algebraic->matrices->L->replaceGlobalValues(global_index,
                                                                            tuple<global_ordinal_type>(global_indices[k]),
                                                                            tuple<scalar>(scalar(0.0)));

                            // Put one on the diagonal element
                            algebraic->matrices->L->replaceGlobalValues(global_index,
                                                                        tuple<global_ordinal_type>(global_index),
                                                                        tuple<scalar>(scalar(1.0)));
                        }
                        else
                        {
                            size_t numColInds = algebraic->graph->K->getNumEntriesInGlobalRow(global_index_r0);
                            global_indices.resize(numColInds);
                            algebraic->graph->K->getGlobalRowCopy(global_index_r0, global_indices(), numColInds);

                            // Replace the initialized elements by zero
                            for (auto k = 0; k < numColInds; ++k)
                                algebraic->matrices->K->replaceGlobalValues(global_index_r0,
                                                                            tuple<global_ordinal_type>(global_indices[k]),
                                                                            tuple<scalar>(scalar(0.0)));

                            // Put one on the diagonal element
                            algebraic->matrices->K->replaceGlobalValues(global_index_r0,
                                                                        tuple<global_ordinal_type>(global_index_r0),
                                                                        tuple<scalar>(scalar(1.0)));

                            if (numPrimalDPN == 4)
                            {
                                size_t numColInds = algebraic->graph->S->getNumEntriesInGlobalRow(global_index_r0);
                                global_indices.resize(numColInds);
                                algebraic->graph->S->getGlobalRowCopy(global_index_r0, global_indices(), numColInds);

                                // Replace the initialized elements by zero
                                for (auto k = 0; k < numColInds; ++k)
                                    algebraic->matrices->S->replaceGlobalValues(global_index_r0,
                                                                                tuple<global_ordinal_type>(global_indices[k]),
                                                                                tuple<scalar>(scalar(0.0)));
                            }
                        }

                        // Put the corresponding value in the right hand side
                        algebraic->vectors->b->replaceGlobalValue(global_index,
                                                                  (scalar(loads->dirichletList->getDirichletValue(i, j))));
                        algebraic->vectors->x->replaceGlobalValue(global_index,
                                                                  (scalar(loads->dirichletList->getDirichletValue(i, j))));
                    }
                }
        }
    }
    if (numPrimalDPN == 4 || numPrimalDPN == 3)
        algebraic->matrices->K->fillComplete();
    if (numPrimalDPN == 4 || numPrimalDPN == 1)
        algebraic->matrices->L->fillComplete();
    if (numPrimalDPN == 4)
        algebraic->matrices->S->fillComplete();

    /*
    if (scaled)
    {
        tpetra_vector_type diag(algebraic->matrices->A->getRowMap());
        algebraic->matrices->A->getLocalDiagCopy(diag);

        Teuchos::ArrayRCP<scalar> diag_entries;
        diag_entries = diag.getDataNonConst();
        for (size_t i = 0; i < diag_entries.size(); ++i)
            diag_entries[i] = 1. / diag_entries[i];

        Teuchos::ArrayRCP<scalar> rhs_entries;
        rhs_entries = algebraic->vectors->b->getDataNonConst();
        for (size_t i = 0; i < rhs_entries.size(); ++i)
            rhs_entries[i] = rhs_entries[i] * diag_entries[i];

        algebraic->matrices->A->leftScale(diag);

        if (domain->contactsList->getContactNumber() >= 1)
            algebraic->matrices->B_T->leftScale(diag);
    }
    */
    // ------------------------------------------
    //
    //       weights computation
    //
    // ------------------------------------------

    auto numMyWeightRegions = loads->weightsList->getWeightRegionsNumber();

    for (auto i = 0; i < numMyWeightRegions; ++i)
    {
        auto currentWeightRegionSize = loads->weightsList->getWeightRegionSize(i);
        Kokkos::parallel_for(
            currentWeightRegionSize, KOKKOS_LAMBDA(const int i_node) {
                global_ordinal_type n1 = loads->weightsList->getNode(i, i_node);

                if (algebraic->map->mapNodes->isNodeGlobalElement(n1))
                {
                    for (auto j = 0; j < numPrimalDPN; ++j)
                    {
                        if (loads->weightsList->getWeightDof(i, j))
                        {
                            global_ordinal_type global_index;

                            if (j == 3)
                            {
                                global_index = n1;
                            }
                            else
                            {
                                if (numPrimalDPN == 4)
                                    global_index = numNodes + 3 * n1 + j;
                                else
                                    global_index = 3 * n1 + j;
                            }

                            algebraic->vectors->weights->sumIntoGlobalValue(global_index,
                                                                            (scalar(loads->weightsList->getWeightValue(i, j))),
                                                                            true);
                        }
                    }
                }
            });
    }

    delete elementVectors_TRI3;
    delete elementVectors_QUAD4;
    delete elementVectors_TETRA4;
    delete elementVectors_HEX8;
}
}; // namespace katoptron
