/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wLine2.h"
#include "wSfLine2.h"
#include "wGaussLine2.h"
#include "wCacheLine2.h"
#include "wMemLine2.h"
#include "wNode.h"
#include "wFct0.h"
#include "wFct1.h"

using namespace tbox;

Line2::Line2(int n, Tag *_ptag, Tag *_etag,
             std::vector<Tag *> &_parts, std::vector<Node *> &nods) : Element(n, _ptag, _etag, _parts, nods)
{
    pmem = new MemLine2(*this, static_cast<int>(getCache().gauss.getN()));
}
Line2::~Line2()
{
    if (pmem)
        delete pmem;
}

/**
 * @brief Return private memory handling Jacobian for this Line2 element
 */
Mem &Line2::getVMem()
{
    return getMem();
}

/**
 * @brief Return cache holding shape functions and Gauss points for all Line2 elements
 */
Cache &Line2::getVCache()
{
    return getCache();
}

/**
 * @brief Compute element unit normal vector
 */
Eigen::Vector3d Line2::normal()
{
    Eigen::Vector3d dx = (nodes[1]->pos - nodes[0]->pos).normalized();
    return Eigen::Vector3d(-dx(1), dx(0), 0.);
}

/**
 * @brief Compute Jacombian determinant at Gauss point k
 * 
 * detJ = norm(x_i * dN_i)
 */
double Line2::buildJ(size_t k) const
{
    Eigen::MatrixXd const &dff = getCache().getDsf(k);

    Eigen::Vector3d t0 = Eigen::Vector3d::Zero();
    size_t i = 0;
    for (auto it = nodes.begin(); it != nodes.end(); ++it, ++i)
        t0 += (*it)->pos * dff(0, i);
    return t0.norm();
}

/**
 * @brief Compute flux vector
 * 
 * s_i(2) = int f * N_i dV 
 */
Eigen::VectorXd Line2::builds(std::vector<double> const &u, Fct0 const &f)
{
    CacheLine2 &cache = getCache();
    GaussLine2 &gauss = cache.gauss;
    MemLine2 &mem = getMem();

    // Gauss integration
    Eigen::VectorXd s = Eigen::Vector2d::Zero();
    for (size_t k = 0; k < gauss.getN(); ++k)
        s += cache.getSf(k) * f.eval(*this, u, k) * gauss.getW(k) * mem.getDetJ(k);
    return s;
}

/**
 * @brief Compute flux vector
 * 
 * s_i(2) = int f_j*n_j * N_i dV 
 */
Eigen::VectorXd Line2::builds(std::vector<double> const &u, Fct1 const &f)
{
    CacheLine2 &cache = getCache();
    GaussLine2 &gauss = cache.gauss;
    MemLine2 &mem = getMem();

    // Gauss integration
    Eigen::VectorXd s = Eigen::Vector2d::Zero();
    for (size_t k = 0; k < gauss.getN(); ++k)
        s += cache.getSf(k) * (f.eval(*this, u, k).dot(normal())) * gauss.getW(k) * mem.getDetJ(k);
    return s;
}

/**
 * @brief Compute square of gradient along element
 * 
 * grad = (Delta_U / Delta_L)^2
 */
double Line2::computeGrad2(std::vector<double> const &u, size_t k)
{
    double dUdL = (u[nodes[1]->row] - u[nodes[0]->row]) / computeV(); // V = length of the element
    return dUdL * dUdL;
}

/**
 * @brief Compute intergal of scalar function
 * 
 * I = int f dV
 */
double Line2::computeV(std::vector<double> const &u, Fct0 const &fct)
{
    CacheLine2 &cache = getCache();
    GaussLine2 &gauss = cache.gauss;
    MemLine2 &mem = getMem();

    // Gauss integration
    double V = 0.0;
    for (size_t k = 0; k < gauss.getN(); ++k)
        V += fct.eval(*this, u, k) * gauss.getW(k) * mem.getDetJ(k);
    return V;
}

/**
 * @brief Return the element length
 */
double Line2::computeV()
{
    return getMem().getVol();
}
